"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
import logging
import numpy
from onnx import TensorProto
from onnx.helper import make_attribute
from onnx.numpy_helper import from_array, to_array
from onnx.defs import onnx_opset_version
from skl2onnx.operator_converters.decision_tree import (
    convert_sklearn_decision_tree_regressor,
    convert_sklearn_decision_tree_classifier)
from skl2onnx.operator_converters.gradient_boosting import (
    convert_sklearn_gradient_boosting_regressor,
    convert_sklearn_gradient_boosting_classifier)
from skl2onnx.operator_converters.random_forest import (
    convert_sklearn_random_forest_classifier,
    convert_sklearn_random_forest_regressor_converter)
from skl2onnx.common.data_types import (
    guess_numpy_type, FloatTensorType, DoubleTensorType)


logger = logging.getLogger('mlprodict.onnx_conv')


def _op_type_domain_regressor(dtype, opsetml):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if opsetml is None:
        from ... import __max_supported_opsets__
        if onnx_opset_version() >= 16:
            opsetml = min(3, __max_supported_opsets__['ai.onnx.ml'])
        else:
            opsetml = min(1, __max_supported_opsets__['ai.onnx.ml'])
    if opsetml >= 3:
        return 'TreeEnsembleRegressor', 'ai.onnx.ml', 3
    if dtype == numpy.float32:
        return 'TreeEnsembleRegressor', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'TreeEnsembleRegressorDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        f"Unsupported dtype {dtype}.")


def _op_type_domain_classifier(dtype, opsetml):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if opsetml >= 3:
        return 'TreeEnsembleClassifier', 'ai.onnx.ml', 3
    if dtype == numpy.float32:
        return 'TreeEnsembleClassifier', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'TreeEnsembleClassifierDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        f"Unsupported dtype {dtype}.")


def _fix_tree_ensemble_node(scope, container, opsetml, node, dtype):
    """
    Fixes a node for old versionsof skl2onnx.
    """
    atts = {'base_values': 'base_values_as_tensor',
            'nodes_hitrates': 'nodes_hitrates_as_tensor',
            'nodes_values': 'nodes_values_as_tensor',
            'target_weights': 'target_weights_as_tensor',
            'class_weights': 'class_weights_as_tensor'}
    logger.debug('postprocess %r name=%r opsetml=%r dtype=%r',
                 node.op_type, node.name, opsetml, dtype)
    if dtype == numpy.float64:
        # Inserting a cast operator.
        index = 0 if node.op_type == 'TreeEnsembleRegressor' else 1
        new_name = scope.get_unique_variable_name('tree_ensemble_cast')
        old_name = node.output[index]
        node.output[index] = new_name
        container.add_node(
            'Cast', [new_name], [old_name], to=TensorProto.DOUBLE,  # pylint: disable=E1101
            name=scope.get_unique_operator_name('tree_ensemble_cast'))
    attributes = list(node.attribute)
    del node.attribute[:]
    for att in attributes:
        if att.name in atts:
            logger.debug('+ rewrite att %r into %r', att.name, atts[att.name])
            if att.type == 6:
                value = from_array(
                    numpy.array(att.floats, dtype=dtype), atts[att.name])
            elif att.type == 4:
                value = from_array(
                    numpy.array(att.t.double_data, dtype=dtype), atts[att.name])
            else:
                raise NotImplementedError(
                    "Unable to postprocess attribute name=%r type=%r "
                    "opsetml=%r op_type=%r (value=%r)." % (
                        att.name, att.type, opsetml, node.op_type, att))
            if to_array(value).shape[0] == 0:
                raise RuntimeError(
                    f"Null value from attribute (dtype={dtype!r}): {att!r}.")
            node.attribute.append(make_attribute(atts[att.name], value))
        else:
            node.attribute.append(att)


def _fix_tree_ensemble(scope, container, opsetml, dtype):
    if opsetml is None:
        from ... import __max_supported_opsets__
        if onnx_opset_version() >= 16:
            opsetml = min(3, __max_supported_opsets__['ai.onnx.ml'])
        else:
            opsetml = min(1, __max_supported_opsets__['ai.onnx.ml'])
    if opsetml < 3 or dtype == numpy.float32:
        return False
    for node in container.nodes:
        if node.op_type not in {'TreeEnsembleRegressor', 'TreeEnsembleClassifier'}:
            continue
        _fix_tree_ensemble_node(scope, container, opsetml, node, dtype)
    container.node_domain_version_pair_sets.add(('ai.onnx.ml', opsetml))
    return True


def new_convert_sklearn_decision_tree_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    if opsetml is None:
        opsetml = 3 if container.target_opset >= 16 else 1
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype, opsetml)
    convert_sklearn_decision_tree_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)


def new_convert_sklearn_decision_tree_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype, opsetml)
    convert_sklearn_decision_tree_regressor(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)


def new_convert_sklearn_gradient_boosting_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    if opsetml is None:
        opsetml = 3 if container.target_opset >= 16 else 1
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype, opsetml)
    convert_sklearn_gradient_boosting_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)


def new_convert_sklearn_gradient_boosting_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype, opsetml)
    convert_sklearn_gradient_boosting_regressor(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)


def new_convert_sklearn_random_forest_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    if (dtype == numpy.float64 and
            isinstance(operator.outputs[1].type, FloatTensorType)):
        operator.outputs[1].type = DoubleTensorType(
            operator.outputs[1].type.shape)
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    if opsetml is None:
        opsetml = 3 if container.target_opset >= 16 else 1
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype, opsetml)
    convert_sklearn_random_forest_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)


def new_convert_sklearn_random_forest_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    if opsetml is None:
        opsetml = 3 if container.target_opset >= 16 else 1
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype, opsetml)
    convert_sklearn_random_forest_regressor_converter(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)
    _fix_tree_ensemble(scope, container, opsetml, dtype)



"""
@file
@brief Direct calls to libraries :epkg:`BLAS` and :epkg:`LAPACK`.
"""
import numpy
from scipy.linalg.blas import sgemm, dgemm  # pylint: disable=E0611
from .direct_blas_lapack import (  # pylint: disable=E0401,E0611
    dgemm_dot, sgemm_dot)


def pygemm(transA, transB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc):
    """
    Pure python implementatin of GEMM.
    """
    if len(A.shape) != 1:
        raise ValueError(  # pragma: no cover
            "A must be a vector.")
    if len(B.shape) != 1:
        raise ValueError(  # pragma: no cover
            "B must be a vector.")
    if len(C.shape) != 1:
        raise ValueError(
            "C must be a vector.")
    if A.shape[0] != M * K:
        raise ValueError(
            f"Dimension mismatch for A.shape={A.shape!r} M={M!r} N={N!r} K={K!r}.")
    if B.shape[0] != N * K:
        raise ValueError(
            f"Dimension mismatch for B.shape={B.shape!r} M={M!r} N={N!r} K={K!r}.")
    if C.shape[0] != N * M:
        raise ValueError(  # pragma: no cover
            f"Dimension mismatch for C.shape={C.shape!r} M={M!r} N={N!r} K={K!r}.")

    if transA:
        a_i_stride = lda
        a_k_stride = 1
    else:
        a_i_stride = 1
        a_k_stride = lda

    if transB:
        b_j_stride = 1
        b_k_stride = ldb
    else:
        b_j_stride = ldb
        b_k_stride = 1

    c_i_stride = 1
    c_j_stride = ldc

    n_loop = 0
    for j in range(N):
        for i in range(M):
            total = 0
            for k in range(K):
                n_loop += 1
                a_index = i * a_i_stride + k * a_k_stride
                if a_index >= A.shape[0]:
                    raise IndexError(
                        "A: i=%d a_index=%d >= %d "
                        "(a_i_stride=%d a_k_stride=%d)" % (
                            i, a_index, A.shape[0], a_i_stride, a_k_stride))
                a_val = A[a_index]

                b_index = j * b_j_stride + k * b_k_stride
                if b_index >= B.shape[0]:
                    raise IndexError(
                        "B: j=%d b_index=%d >= %d "
                        "(a_i_stride=%d a_k_stride=%d)" % (
                            j, b_index, B.shape[0], b_j_stride, b_k_stride))
                b_val = B[b_index]

                mult = a_val * b_val
                total += mult

            c_index = i * c_i_stride + j * c_j_stride
            if c_index >= C.shape[0]:
                raise IndexError("C: %d >= %d" % (c_index, C.shape[0]))
            C[c_index] = alpha * total + beta * C[c_index]

    if n_loop != M * N * K:
        raise RuntimeError(  # pragma: no cover
            "Unexpected number of loops: %d != %d = (%d * %d * %d) "
            "lda=%d ldb=%d ldc=%d" % (
                n_loop, M * N * K, M, N, K, lda, ldb, ldc))


def gemm_dot(A, B, transA=False, transB=False):
    """
    Implements dot product with gemm when possible.

    :param A: first matrix
    :param B: second matrix
    :param transA: is first matrix transposed?
    :param transB: is second matrix transposed?
    """
    if A.dtype != B.dtype:
        raise TypeError(  # pragma: no cover
            f"Matrices A and B must have the same dtype not {A.dtype!r}, {B.dtype!r}.")
    if len(A.shape) != 2:
        raise ValueError(  # pragma: no cover
            f"Matrix A does not have 2 dimensions but {len(A.shape)}.")
    if len(B.shape) != 2:
        raise ValueError(  # pragma: no cover
            f"Matrix B does not have 2 dimensions but {len(B.shape)}.")

    def _make_contiguous_(A, B):
        if not A.flags['C_CONTIGUOUS']:
            A = numpy.ascontiguousarray(A)
        if not B.flags['C_CONTIGUOUS']:
            B = numpy.ascontiguousarray(B)
        return A, B

    all_dims = A.shape + B.shape
    square = min(all_dims) == max(all_dims)

    if transA:
        if transB:
            if A.dtype == numpy.float32:
                if square:
                    C = numpy.zeros((A.shape[1], B.shape[0]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    sgemm_dot(B, A, True, True, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[1], B.shape[0]), dtype=A.dtype)
                    return sgemm(1, A, B, 0, C, 1, 1, 1)
            if A.dtype == numpy.float64:
                if square:
                    C = numpy.zeros((A.shape[1], B.shape[0]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    dgemm_dot(B, A, True, True, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[1], B.shape[0]), dtype=A.dtype)
                    return dgemm(1, A, B, 0, C, 1, 1, 1)
            return A.T @ B.T
        else:
            if A.dtype == numpy.float32:
                if square:
                    C = numpy.zeros((A.shape[1], B.shape[1]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    sgemm_dot(B, A, False, True, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[1], B.shape[1]), dtype=A.dtype)
                    return sgemm(1, A, B, 0, C, 1, 0, 1)
            if A.dtype == numpy.float64:
                if square:
                    C = numpy.zeros((A.shape[1], B.shape[1]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    dgemm_dot(B, A, False, True, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[1], B.shape[1]), dtype=A.dtype)
                    return dgemm(1, A, B, 0, C, 1, 0, 1)
            return A.T @ B
    else:
        if transB:
            if A.dtype == numpy.float32:
                if square:
                    C = numpy.zeros((A.shape[0], B.shape[0]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    sgemm_dot(B, A, True, False, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[0], B.shape[0]), dtype=A.dtype)
                    return sgemm(1, A, B, 0, C, 0, 1, 1)
            if A.dtype == numpy.float64:
                if square:
                    C = numpy.zeros((A.shape[0], B.shape[0]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    dgemm_dot(B, A, True, False, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[0], B.shape[0]), dtype=A.dtype)
                    return dgemm(1, A, B, 0, C, 0, 1, 1)
            return A @ B.T
        else:
            if A.dtype == numpy.float32:
                if square:
                    C = numpy.zeros((A.shape[0], B.shape[1]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    sgemm_dot(B, A, False, False, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[0], B.shape[1]), dtype=A.dtype)
                    return sgemm(1, A, B, 0, C, 0, 0)
            if A.dtype == numpy.float64:
                if square:
                    C = numpy.zeros((A.shape[0], B.shape[1]), dtype=A.dtype)
                    A, B = _make_contiguous_(A, B)
                    dgemm_dot(B, A, False, False, C)
                    return C
                else:
                    C = numpy.zeros((A.shape[0], B.shape[1]), dtype=A.dtype)
                    return dgemm(1, A, B, 0, C, 0, 0, 1)
            return A @ B



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class DepthToSpace(OpRun):

    atts = {'blocksize': 0, 'mode': b'DCR'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=DepthToSpace.atts,
                       **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data.shape) != 4:
            raise RuntimeError(  # pragma: no cover
                f"Unexpected shape {data.shape!r}.")
        b, c, h, w = data.shape
        if self.mode == b'DCR':
            tmpshape = (b, self.blocksize, self.blocksize,
                        c // (self.blocksize * self.blocksize), h, w)
            reshaped = data.reshape(tmpshape)
            transposed = numpy.transpose(reshaped, [0, 3, 4, 1, 5, 2])
        else:
            # assert mode == "CRD"
            tmpshape = (b, c // (self.blocksize * self.blocksize),
                        self.blocksize, self.blocksize, h, w)
            reshaped = data.reshape(tmpshape)
            transposed = numpy.transpose(reshaped, [0, 1, 4, 2, 5, 3])
        finalshape = (b, c // (self.blocksize * self.blocksize),
                      h * self.blocksize, w * self.blocksize)
        y = numpy.reshape(transposed, finalshape)
        return (y, )


class SpaceToDepth(OpRun):

    atts = {'blocksize': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=SpaceToDepth.atts,
                       **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data.shape) != 4:
            raise RuntimeError(  # pragma: no cover
                f"Unexpected shape {data.shape!r}.")
        b, C, H, W = data.shape
        tmpshape = (b, C, H // self.blocksize, self.blocksize,
                    W // self.blocksize, self.blocksize)
        reshaped = numpy.reshape(data, tmpshape)
        transposed = numpy.transpose(reshaped, [0, 3, 5, 1, 2, 4])
        finalshape = (b, C * self.blocksize * self.blocksize,
                      H // self.blocksize, W // self.blocksize)
        y = numpy.reshape(transposed, finalshape)
        return (y, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRun
from ._op_onnx_numpy import (  # pylint: disable=E0611,E0401
    topk_element_min_double, topk_element_max_double, topk_element_fetch_double,
    topk_element_min_float, topk_element_max_float, topk_element_fetch_float,
    topk_element_min_int64, topk_element_max_int64, topk_element_fetch_int64)


def topk_sorted_implementation(X, k, axis, largest):
    """
    Retrieves the top-k elements.

    @param      X           data
    @param      k           k in top-k
    @param      axis        axis chosen to select the top-k elements
    @param      largest     largest (1) or smallest (0)
    @return                 top-k values, top-k indices

    See function `_kneighbors_reduce_func
    <https://github.com/scikit-learn/scikit-learn/tree/master/
    sklearn/neighbors/base.py#L304>`_.
    """
    if isinstance(k, numpy.ndarray):
        if k.size != 1:
            raise RuntimeError(  # pragma: no cover
                f"k must be an integer not {k!r}.")
        k = k[0]
    if len(X.shape) == 2 and axis == 1:
        sample_range = numpy.arange(X.shape[0])[:, None]
        if largest == 0:
            sorted_indices = numpy.argpartition(X, axis=axis, kth=k - 1)
            sorted_indices = sorted_indices[:, :k]
            # argpartition doesn't guarantee sorted order, so we sort again
            sorted_indices = sorted_indices[
                sample_range, numpy.argsort(X[sample_range, sorted_indices])]
        else:
            sorted_indices = numpy.argpartition(-X, axis=axis, kth=k - 1)
            sorted_indices = sorted_indices[:, :k]
            # argpartition doesn't guarantee sorted order, so we sort again
            sorted_indices = sorted_indices[
                sample_range, numpy.argsort(-X[sample_range, sorted_indices])]
        sorted_distances = X[sample_range, sorted_indices]
        return sorted_distances, sorted_indices

    sorted_indices = numpy.argsort(X, axis=axis)
    sorted_values = numpy.sort(X, axis=axis)
    if largest:
        sorted_indices = numpy.flip(sorted_indices, axis=axis)
        sorted_values = numpy.flip(sorted_values, axis=axis)
    ark = numpy.arange(k)
    topk_sorted_indices = numpy.take(sorted_indices, ark, axis=axis)
    topk_sorted_values = numpy.take(sorted_values, ark, axis=axis)
    return topk_sorted_values, topk_sorted_indices


def topk_sorted_implementation_cpp(X, k, axis, largest, th_para=50):
    """
    Retrieves the top-k elements using a C++
    implementation when the axis is the last dimension,
    otherwise, it falls back to
    @see fn topk_sorted_implementation.

    @param      X           data
    @param      k           k in top-k
    @param      axis        axis chosen to select the top-k elements
    @param      largest     largest (1) or smallest (0)
    @param      th_para     threshold for parallelisation
    @return                 top-k values, top-k indices
    """
    if isinstance(k, numpy.ndarray):
        if k.size != 1:
            raise RuntimeError(  # pragma: no cover
                f"k must be an integer not {k!r}.")
    if axis != len(X.shape) - 1:
        if k == 0:
            return numpy.empty((0,), dtype=numpy.int64)
        return topk_sorted_implementation(X, k, axis, largest)
    if X.dtype == numpy.float64:
        if k == 0:
            return numpy.empty((0,), dtype=X.dtype), numpy.empty((0,), dtype=numpy.int64)
        if largest:
            topk_sorted_indices = topk_element_max_double(X, k, True, th_para)
        else:
            topk_sorted_indices = topk_element_min_double(X, k, True, th_para)
        topk_sorted_values = topk_element_fetch_double(X, topk_sorted_indices)
    elif X.dtype == numpy.float32:
        if k == 0:
            return numpy.empty((0,), dtype=X.dtype), numpy.empty((0,), dtype=numpy.int64)
        if largest:
            topk_sorted_indices = topk_element_max_float(X, k, True, th_para)
        else:
            topk_sorted_indices = topk_element_min_float(X, k, True, th_para)
        topk_sorted_values = topk_element_fetch_float(X, topk_sorted_indices)
    elif X.dtype == numpy.int64:
        if k == 0:
            return numpy.empty((0,), dtype=X.dtype), numpy.empty((0,), dtype=numpy.int64)
        if largest:
            topk_sorted_indices = topk_element_max_int64(X, k, True, th_para)
        else:
            topk_sorted_indices = topk_element_min_int64(X, k, True, th_para)
        topk_sorted_values = topk_element_fetch_int64(X, topk_sorted_indices)
    else:
        if k == 0:
            return numpy.empty((0,), dtype=numpy.int64)
        return topk_sorted_implementation(X, k, axis, largest)
    return topk_sorted_values, topk_sorted_indices


class _CommonTopK(OpRun):
    """
    Ths class hides a parameter used as a threshold above
    which the parallelisation is started: ``th_para``.
    """

    atts = {'axis': -1}

    def __init__(self, *args, **options):
        OpRun.__init__(self, *args, **options)
        self.th_para = 50

    def _common_run(self, data, ink, largest=1):  # pylint: disable=W0221
        """
        Runtime for operator *TopK*.
        The implementation is not the most efficient
        as it sorts everything then extracts the top *k*
        values.

        .. warning::
            ONNX specifications may be imprecise in case of negative value
            for axis. The implementation follows what :epkg:`onnxruntime`
            does in `top_k.cc
            <https://github.com/Microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/math/top_k.cc#L63>`_.
        """
        k = ink[0]
        axis = self.axis if self.axis >= 0 else (self.axis + len(data.shape))
        sort, sorti = topk_sorted_implementation_cpp(
            data, k, axis, largest, self.th_para)
        return (sort, sorti.astype(numpy.int64))


class TopK_1(_CommonTopK):

    atts = {'axis': -1, 'k': None}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonTopK.__init__(self, onnx_node, desc=desc,
                             expected_attributes=TopK_10.atts,
                             **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        Runtime for operator *TopK*.
        The implementation is not the most efficient
        as it sorts everything then extracts the top *k*
        values.

        .. warning::
            ONNX specifications may be imprecise in case of negative value
            for axis. The implementation follows what :epkg:`onnxruntime`
            does in `top_k.cc
            <https://github.com/Microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/math/top_k.cc#L63>`_.
        """
        return _CommonTopK._common_run(self, data, [self.k])


class TopK_10(_CommonTopK):

    atts = {'axis': -1}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonTopK.__init__(self, onnx_node, desc=desc,
                             expected_attributes=TopK_10.atts,
                             **options)

    def _run(self, data, ink, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        Runtime for operator *TopK*.
        The implementation is not the most efficient
        as it sorts everything then extracts the top *k*
        values.

        .. warning::
            ONNX specifications may be imprecise in case of negative value
            for axis. The implementation follows what :epkg:`onnxruntime`
            does in `top_k.cc
            <https://github.com/Microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/math/top_k.cc#L63>`_.
        """
        return _CommonTopK._common_run(self, data, ink)


class TopK_11(_CommonTopK):

    atts = {'axis': -1, 'largest': 1, 'sorted': 1}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonTopK.__init__(self, onnx_node, desc=desc,
                             expected_attributes=TopK_11.atts,
                             **options)
        if self.sorted not in (True, 1):
            raise RuntimeError(  # pragma: no cover
                "TopK does not implement anything for sorted=0.")

    def _run(self, data, ink, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        Runtime for operator *TopK*.
        The implementation is not the most efficient
        as it sorts everything then extracts the top *k*
        values.

        .. warning::
            ONNX specifications may be imprecise in case of negative value
            for axis. The implementation follows what :epkg:`onnxruntime`
            does in `top_k.cc
            <https://github.com/Microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/math/top_k.cc#L63>`_.
        """
        return _CommonTopK._common_run(self, data, ink, self.largest)


if onnx_opset_version() >= 11:
    TopK = TopK_11
elif onnx_opset_version() >= 10:  # pragma: no cover
    TopK = TopK_10
else:  # pragma: no cover
    TopK = TopK_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRunUnaryNum


class Clip_6(OpRunUnaryNum):

    atts = {'min': -3.4028234663852886e+38,
            'max': 3.4028234663852886e+38}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Clip_6.atts,
                               **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and data.flags['WRITEABLE']:
            return self._run_inplace(data)
        res = numpy.clip(data, self.min, self.max)
        return (res, ) if res.dtype == data.dtype else (res.astype(data.dtype), )

    def _run_inplace(self, data):
        return (numpy.clip(data, self.min, self.max, out=data), )

    def to_python(self, inputs):
        return ("import numpy",
                f"return numpy.clip({inputs[0]}, min_, max_)")


class Clip_11(OpRunUnaryNum):

    version_higher_than = 11
    mandatory_inputs = ['X']
    optional_inputs = OrderedDict([
        ('min', -3.4028234663852886e+38),
        ('max', 3.4028234663852886e+38)
    ])

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               **options)

    def run(self, x, *minmax, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202,W0221
        """
        Calls method ``_run``.
        """
        try:
            res = self._run(x, *minmax, attributes=attributes,
                            verbose=verbose, fLOG=fLOG)
        except TypeError as e:  # pragma: no cover
            raise TypeError("Issues with types {} (binary operator {}).".format(
                ", ".join(str(type(_)) for _ in [x]),
                self.__class__.__name__)) from e
        return res

    def _run(self, data, *minmax, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and data.flags['WRITEABLE']:
            return self._run_inplace(data, *minmax)
        le = len(minmax)
        amin = minmax[0] if le > 0 else None  # -3.4028234663852886e+38
        amax = minmax[1] if le > 1 else None  # 3.4028234663852886e+38
        if amin is None and amax is None:
            amin = -numpy.inf
        res = numpy.clip(data, amin, amax)
        return (res, ) if res.dtype == data.dtype else (res.astype(data.dtype), )

    def _run_inplace(self, data, *minmax):  # pylint: disable=W0221
        le = len(minmax)
        amin = minmax[0] if le > 0 else None  # -3.4028234663852886e+38
        amax = minmax[1] if le > 1 else None  # 3.4028234663852886e+38
        res = numpy.clip(data, amin, amax, out=data)
        return (res, )

    def to_python(self, inputs):
        return ("import numpy",
                f"return numpy.clip({inputs[0]}, min_, max_)")


if onnx_opset_version() >= 11:
    Clip = Clip_11
else:
    Clip = Clip_6  # pragma: no cover



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _apply_adam(r, t, x, g, v, h,
                norm_coefficient, norm_coefficient_post,
                alpha, beta, epsilon):  # type: ignore
    # Add gradient of regularization term.
    g_regularized = norm_coefficient * x + g
    # Update momentum.
    v_new = alpha * v + (1 - alpha) * g_regularized
    # Update second-order momentum.
    h_new = beta * h + (1 - beta) * (g_regularized * g_regularized)
    # Compute element-wise square root.
    h_sqrt = numpy.sqrt(h_new) + epsilon
    # Adjust learning rate.
    r_adjusted = None
    if t > 0:
        # Consider bias correction on momentums.
        r_adjusted = r * numpy.sqrt(1 - beta**t) / (1 - alpha**t)
    else:
        # No bias correction on momentums.
        r_adjusted = r
    # Apply Adam update rule.
    x_new = x - r_adjusted * (v_new / h_sqrt)
    # It's possible to apply regularization in the end.
    x_final = (1 - norm_coefficient_post) * x_new
    return x_final, v_new, h_new


class Adam(OpRun):

    atts = {'alpha': 0.8999999761581421,
            'beta': 0.9990000128746033,
            'epsilon': 9.999999974752427e-07,
            'norm_coefficient': 0.,
            'norm_coefficient_post': 0.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Adam.atts,
                       **options)

    def _run(self, *data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data) == 6:
            return self._run1(*data)
        n = (len(data) - 2) // 4
        xs = []
        vs = []
        hs = []
        for i in range(0, n):
            a, b, c = self._run1(*data[:2], data[2 + i],
                                 data[2 + n + i], data[2 + n * 2 + i],
                                 data[2 + n * 3 + i])
            xs.append(a)
            vs.append(b)
            hs.append(c)
        return tuple(xs + vs + hs)

    def _run1(self, r, t, x, g, v, h):  # pylint: disable=W0221
        x_new, v_new, h_new = _apply_adam(
            r, t, x, g, v, h, self.norm_coefficient,
            self.norm_coefficient_post, self.alpha, self.beta, self.epsilon)
        return x_new, v_new, h_new



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class HardSigmoid(OpRunUnaryNum):

    atts = {'alpha': 0.2, 'beta': 0.5}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=HardSigmoid.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x)
        y = numpy.maximum(0, numpy.minimum(1, x * self.alpha + self.beta))
        return (y, )

    def _run_inplace(self, x):
        x *= self.alpha
        x += self.beta
        numpy.minimum(x, 1, out=x)
        numpy.maximum(x, 0, out=x)
        return (x, )

    def to_python(self, inputs):
        return (
            "import numpy",
            "return numpy.maximum(0, numpy.minimum(1, {0} * {1} + {2}))".format(
                inputs[0], self.alpha, self.beta))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunBinaryNumpy


class Div(OpRunBinaryNumpy):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNumpy.__init__(self, numpy.divide, onnx_node,
                                  desc=desc, **options)

    def _run(self, a, b, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = OpRunBinaryNumpy._run(self, a, b)
        if res[0].dtype != a.dtype:
            return (res[0].astype(a.dtype), )
        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRun


class Shape_1(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.array(data.shape, dtype=numpy.int64), )


class Shape_15(Shape_1):

    atts = {'start': 0, 'end': numpy.nan}

    def __init__(self, onnx_node, desc=None, **options):
        Shape_1.__init__(self, onnx_node, desc=desc,
                         expected_attributes=Shape_15.atts, **options)

    def _interval(self, n):
        if self.start == 0:
            if numpy.isnan(self.end):
                return None
            elif self.end < 0:
                return (0, n + self.end)
            return (0, self.end)
        if numpy.isnan(self.end):
            return (self.start, n)
        elif self.end < 0:
            return (self.start, n + self.end)
        return (self.start, self.end)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        ab = self._interval(len(data.shape))
        if ab is None:
            return (numpy.array(data.shape, dtype=numpy.int64), )
        return (numpy.array(data.shape[ab[0]: ab[1]], dtype=numpy.int64), )


if onnx_opset_version() >= 15:
    Shape = Shape_15
else:  # pragma: no cover
    Shape = Shape_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Elu(OpRunUnaryNum):

    atts = {'alpha': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Elu.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.where(x > 0, x, self.alpha * (numpy.exp(x) - 1)), )

    def to_python(self, inputs):
        return (
            "import numpy",
            ("return numpy.where({0} > 0, {0}, "
             "{1} * (numpy.exp({0}) - 1))").format(inputs[0], self.alpha))



"""
@file
@brief Numpy helpers for the conversion from onnx to numpy.
"""
import numpy


def make_slice(data, starts, ends, axes=None, steps=None):
    """
    Implements operator slice in numpy.

    :param data: input
    :param starts: mandatory
    :param ends: mandatory
    :param axes: optional
    :param steps: optional
    :return: results
    """
    slices = [slice(0, data.shape[i]) for i in range(len(data.shape))]
    if axes is None:
        axes = range(len(starts))
    for i, a in enumerate(axes):
        if steps is None:
            slices[a] = slice(starts[i], ends[i])
        else:
            slices[a] = slice(starts[i], ends[i], steps[i])
    return data[slices]


def argmax_use_numpy_select_last_index(
        data, axis=0, keepdims=True, select_last_index=False):
    """
    Needed or operator `ArgMax`.
    """
    if not select_last_index:
        result = numpy.argmax(data, axis=axis)
        if keepdims and len(result.shape) < len(data.shape):
            result = numpy.expand_dims(result, axis)
        return result.astype(numpy.int64)

    data = numpy.flip(data, axis)
    result = numpy.argmax(data, axis=axis)
    result = data.shape[axis] - result - 1
    if keepdims:
        result = numpy.expand_dims(result, axis)
    return result.astype(numpy.int64)


def argmin_use_numpy_select_last_index(
        data, axis=0, keepdims=True, select_last_index=False):
    """
    Needed or operator `ArgMin`.
    """
    if not select_last_index:
        result = numpy.argmin(data, axis=axis)
        if keepdims and len(result.shape) < len(data.shape):
            result = numpy.expand_dims(result, axis)
        return result.astype(numpy.int64)

    data = numpy.flip(data, axis)
    result = numpy.argmin(data, axis=axis)
    result = data.shape[axis] - result - 1
    if keepdims:
        result = numpy.expand_dims(result, axis)
    return result.astype(numpy.int64)


def array_feature_extrator(data, indices):
    """
    Implementation of operator *ArrayFeatureExtractor*
    with :epkg:`numpy`.
    """
    if len(indices.shape) == 2 and indices.shape[0] == 1:
        index = indices.ravel().tolist()
        add = len(index)
    elif len(indices.shape) == 1:
        index = indices.tolist()
        add = len(index)
    else:
        add = 1
        for s in indices.shape:
            add *= s
        index = indices.ravel().tolist()
    if len(data.shape) == 1:
        new_shape = (1, add)
    else:
        new_shape = list(data.shape[:-1]) + [add]
    tem = data[..., index]
    res = tem.reshape(new_shape)
    return res


class NumpyCode:
    """
    Converts an ONNX operators into :epkg:`numpy` code.

    :param opset: target opset for the conversion (usually unused)
    :param name: node name
    :param op_type: operator type
    :param domain: domain
    :param inputs: inputs
    :param outputs: outputs
    :param attributes: attributes
    :param used: dictionary `{k: v}`,
        list of nodes taking *k* as input
    :param context: whole context
    :param mark_inits: marks initializer as replaced
    :param indent: indentation of the second line and following
    :return: code as str
    """

    def __init__(self, opset, name=None, op_type=None, domain='',
                 inputs=None, outputs=None, attributes=None,
                 used=None, context=None, mark_inits=None,
                 indent="", **unused):
        self.opset = opset
        self.name = name
        self.op_type = op_type
        self.domain = domain
        self.inputs = inputs
        self.outputs = outputs
        self.attributes = attributes
        self.used = used
        self.context = context
        self.mark_inits = mark_inits
        self.unused = unused
        self.indent = indent

    def _make_sure_inputs(self, n, m=None):
        if m is None:
            m = n
        if len(self.inputs) < n:
            raise RuntimeError(  # pragma: no cover
                "Expecting at least %d inputs for operator %r not %r." % (
                    n, self.op_type, self.inputs))
        if len(self.inputs) > m:
            raise RuntimeError(  # pragma: no cover
                "Expecting at most %d inputs for operator %r not %r." % (
                    m, self.op_type, self.inputs))

    def _make_sure_opsets(self, mi, ma=None):
        if mi is not None and self.opset < mi:
            raise RuntimeError(  # pragma: no cover
                "Cannot convert operator type %d, opset %d < %d." % (
                    self.op_type, self.opset, mi))
        if ma is not None and self.opset > ma:
            raise RuntimeError(  # pragma: no cover
                "Cannot convert operator type %d, opset %d > %d." % (
                    self.op_type, self.opset, mi))

    def _getat(self, name, defval=None, format=None):

        def f(v):
            if format is None:
                return v
            if format == 'listint' and isinstance(v, str):
                return list(
                    map(int, v.strip('[]').replace(' ', '').split(',')))
            if format == 'listfloat' and isinstance(v, str):
                return list(
                    map(float, v.strip('[]').replace(' ', '').split(',')))
            raise ValueError(  # pragma: no cover
                f"Unable to convert {v!r} with format={format!r}.")

        for n, val in self.attributes:
            if name == n:
                return f(val)
        return defval

    def _simplify(self, name, kind):
        value = None
        if (self.used is not None and name in self.used and
                len(self.used[name]) == 1 and self.context is not None):
            inits = self.context['initializers_dict']
            if name in inits:
                v = inits[name]
                if v.dtype == numpy.int64 and v.size < 10:
                    value = v
                    if name not in self.mark_inits:
                        self.mark_inits[name] = []
                    self.mark_inits[name].append(v)

        if kind == 'tuple':
            if value is None:
                return f"tuple({name})"
            if value.size == 1:
                return str(tuple(value)[0])
            return str(tuple(value))
        elif kind == 'list':
            if value is None:
                return name
            if len(value.shape) == 0:
                return str(value)
            return str(list(value))
        raise NotImplementedError(  # pragma: no cover
            f"Unknown scenario to simplify ({kind!r}).")

    @staticmethod
    def _make_tuple(val):
        if isinstance(val, tuple):
            return val
        if isinstance(val, list):
            return tuple(val)
        if isinstance(val, int):
            return val
        if isinstance(val, str):
            return tuple(map(int, val.strip('()[]').replace(" ", "").split(",")))
        raise NotImplementedError(  # pragma: no cover
            f"Unable to convert type {type(val)!r} ({val!r}) into tuple.")

    def make_numpy_code(self):
        """
        Main method, returns the python code for a given
        operator.
        """
        if self.domain == '':
            return self._make_numpy_code_onnx()

        if self.domain == 'ai.onnx.ml':
            return self._make_numpy_code_onnxml()

        if self.domain == 'com.microsoft':
            return self._make_numpy_code_others()

        raise NotImplementedError(  # pragma: no cover
            f"Unable to convert any operator from domain {self.domain!r}.")

    def _make_numpy_code_onnx(self):

        binary_ops = dict(Add='+', Sub='-', Div='/', Mul='*', MatMul='@',
                          Pow='**')
        unary_ops = dict(Neg='-')
        unary_ops_ = dict(Sqrt='** 0.5')

        outs = ", ".join(self.outputs)

        if self.op_type in binary_ops:
            self._make_sure_inputs(2)
            return "%s = %s %s %s" % (
                outs, self.inputs[0], binary_ops[self.op_type],
                self.inputs[1])

        if self.op_type in unary_ops:
            self._make_sure_inputs(1)
            return f"{outs} = {unary_ops[self.op_type]} {self.inputs[0]}"

        if self.op_type in unary_ops_:
            self._make_sure_inputs(1)
            return f"{outs} = {self.inputs[0]} {unary_ops_[self.op_type]}"

        if self.op_type in {'Abs', 'Ceil', 'Cos', 'Cosh',
                            'Exp', 'Log', 'Sin', 'Sinh',
                            'Tan', 'Tanh'}:
            return f"{outs} = numpy.{self.op_type.lower()}({self.inputs[0]})"

        if self.op_type == 'ArgMax':
            self._make_sure_opsets(12)
            self._make_sure_inputs(1)
            axis = self._getat('axis', 0)
            keepdims = self._getat('keepdims', 1)
            select_last_index = self._getat('keepdims', 0)
            if select_last_index:
                return (
                    "%s = argmax_use_numpy_select_last_index("
                    "%s, axis=%s, keepdims=%s, select_last_index=%s)" % (
                        outs, self.inputs[0], axis, keepdims, select_last_index))
            if keepdims:
                return "%s = numpy.expand_dims(numpy.argmax(%s, axis=%s), -1)" % (
                    outs, self.inputs[0], axis)
            return f"{outs} = numpy.argmax({self.inputs[0]}, axis={axis})"

        if self.op_type == 'ArgMin':
            self._make_sure_opsets(12)
            self._make_sure_inputs(1)
            axis = self._getat('axis', 0)
            keepdims = self._getat('keepdims', 1)
            select_last_index = self._getat('keepdims', 0)
            if select_last_index:
                return (
                    "%s = argmin_use_numpy_select_last_index("
                    "%s, axis=%s, keepdims=%s, select_last_index=%s)" % (
                        outs, self.inputs[0], axis, keepdims, select_last_index))
            if keepdims:
                return "%s = numpy.expand_dims(numpy.argmin(%s, axis=%s), -1)" % (
                    outs, self.inputs[0], axis)
            return f"{outs} = numpy.argmin({self.inputs[0]}, axis={axis})"

        if self.op_type == 'Cast':
            from ..onnx2py_helper import _elem_type_as_str
            self._make_sure_inputs(1)
            to = int(self._getat('to', 1))
            dtype = _elem_type_as_str(to)
            dtype = {'double': 'float64', 'float': 'float32'}.get(dtype, dtype)
            return f"{outs} = {self.inputs[0]}.astype(numpy.{dtype})"

        if self.op_type == 'Concat':
            axis = self._getat('axis', 0)
            return f"{outs} = numpy.concatenate([{', '.join(self.inputs)}], {axis})"

        if self.op_type == 'ConstantOfShape':
            self._make_sure_opsets(9)
            self._make_sure_inputs(1)
            value = self._getat('value', 0, format='listfloat')
            shape = self._simplify(self.inputs[0], kind='tuple')
            return f"{outs} = numpy.full({shape}, {value})"

        if self.op_type == 'Max':
            return f"{outs} = numpy.maximum({', '.join(self.inputs)})"

        if self.op_type == 'Gather':
            self._make_sure_opsets(11)
            self._make_sure_inputs(2)
            axis = self._getat('axis', 0)
            return "%s = numpy.take(%s, %s, axis=%s)" % (
                outs, self.inputs[0],
                self._simplify(self.inputs[1], 'list'), axis)

        if self.op_type == 'Gemm':
            self._make_sure_inputs(2, 3)
            alpha = self._getat('alpha', 0.)
            transA = self._getat('transA', 0)
            transB = self._getat('transB', 0)
            ta = ".T" if transA in ('1', 1, True) else ""
            tb = ".T" if transB in ('1', 1, True) else ""
            if len(self.inputs) == 2:
                return f"{outs} = {self.inputs[0]}{ta} @ {self.inputs[1]}{tb} * {alpha}"
            beta = self._getat('beta', 0.)
            return "%s = %s%s @ %s%s * %s + %s * %s" % (
                outs, self.inputs[0], ta, self.inputs[1], tb, alpha,
                self.inputs[2], beta)

        if self.op_type == 'Identity':
            return f"{outs} = {self.inputs[0]}"

        if self.op_type == 'ReduceProd':
            self._make_sure_inputs(1)
            axes = self._getat('axes', "[0]")
            keepdims = self._getat('keepdims', 0)
            return "%s = %s.prod(axis=tuple(%s), keepdims=%s)" % (
                outs, self.inputs[0], axes, keepdims)

        if self.op_type == 'ReduceSum':
            self._make_sure_opsets(11)
            self._make_sure_inputs(2)
            keepdims = self._getat('keepdims', 0)
            return "%s = %s.sum(axis=%s, keepdims=%s)" % (
                outs, self.inputs[0], self._simplify(self.inputs[1], 'tuple'),
                keepdims)

        if self.op_type == 'ReduceSumSquare':
            self._make_sure_inputs(1)
            axes = self._getat('axes', "[0]")
            keepdims = self._getat('keepdims', 0)
            return "%s = (%s ** 2).sum(axis=tuple(%s), keepdims=%s)" % (
                outs, self.inputs[0], axes, keepdims)

        if self.op_type == 'Reshape':
            self._make_sure_inputs(2)
            simp = self._simplify(self.inputs[1], 'tuple')
            return f"{outs} = {self.inputs[0]}.reshape({simp})"

        if self.op_type == 'Shape':
            self._make_sure_inputs(1)
            return f"{outs} = numpy.array({self.inputs[0]}.shape, dtype=numpy.int64)"

        if self.op_type == 'Slice':
            return f"{outs} = make_slice({', '.join(self.inputs)})"

        if self.op_type == 'Softmax':
            self._make_sure_inputs(1)
            axis = self._getat('axis', -1)
            return f"{outs} = scipy_special.softmax({self.inputs[0]}, axis={axis})"

        if self.op_type == 'Squeeze':
            self._make_sure_opsets(13)
            self._make_sure_inputs(2)
            return "%s = numpy.squeeze(%s, axis=%s)" % (
                outs, self.inputs[0], self._simplify(self.inputs[1], 'tuple'))

        if self.op_type == 'Transpose':
            self._make_sure_inputs(1)
            perm = self._getat('perm', None)
            return "%s = numpy.transpose(%s, axes=%s)" % (
                outs, self.inputs[0], self._make_tuple(perm))

        if self.op_type == 'Unsqueeze':
            self._make_sure_opsets(13)
            self._make_sure_inputs(2)
            return "%s = numpy.expand_dims(%s, axis=%s)" % (
                outs, self.inputs[0],
                self._simplify(self.inputs[1], 'tuple'))

        raise NotImplementedError(  # pragma: no cover
            f"Unable to convert operator type {self.op_type!r} name={self.name!r}.")

    def _make_numpy_code_onnxml(self):
        outs = ", ".join(self.outputs)

        if self.op_type == 'ArrayFeatureExtractor':
            self._make_sure_inputs(2)
            return "%s = array_feature_extrator(%s, %s)" % (
                outs, self.inputs[0], self.inputs[1])

        if self.op_type == 'LinearClassifier':
            multi_class = self._getat('targets', 0)
            if multi_class != 0:
                raise NotImplementedError(  # pragma: no cover
                    "Conversion of operator %r with multi_class=%r "
                    "is not implemented." % (self.op_type, multi_class))
            self._make_sure_inputs(1)
            coefficients = self._getat('coefficients', None)
            intercepts = self._getat('intercepts', None)
            post_transform = self._getat(
                'post_transform', 'NONE').strip('"\'b')
            classlabels_strings = self._getat('classlabels_strings', None)
            if classlabels_strings is not None:
                raise NotImplementedError(  # pragma: no cover
                    "Conversion of operator %r with classlabels_strings=%r "
                    "is not implemented." % (self.op_type, classlabels_strings))
            classlabels_ints = self._getat(
                'classlabels_ints', None, format="listint")
            if classlabels_ints != list(range(len(classlabels_ints))):
                raise NotImplementedError(  # pragma: no cover
                    "Conversion of operator %r with classlabels_ints=%r!=%r "
                    "is not implemented." % (
                        self.op_type, classlabels_ints,
                        list(range(len(classlabels_ints)))))
            targets = len(classlabels_ints)
            rows = [
                "coefs = numpy.array(%s, dtype=numpy.float32)."
                "reshape((%d, -1)).T" % (coefficients, targets),
                "%sinter = numpy.array(%s, dtype=numpy.float32)."
                "reshape((-1, %d))" % (self.indent, intercepts, targets)]

            if post_transform == "SOFTMAX":
                rows.append(
                    "%s%s = scipy_special.softmax"
                    "(%s @ coefs + inter, axis=1)" % (
                        self.indent, self.outputs[1], self.inputs[0]))
            elif post_transform == 'NONE':
                rows.append(
                    "%s%s = %s @ coefs + inter" % (
                        self.indent, self.outputs[1], self.inputs[0]))
            elif post_transform != "NONE":
                raise NotImplementedError(  # pragma: no cover
                    "Conversion of operator %r with post_transform=%r "
                    "is not implemented." % (self.op_type, post_transform))
            rows.append("%s%s = numpy.argmax(%s, axis=1)" % (
                self.indent, self.outputs[0], self.outputs[1]))
            return "\n".join(rows)

        if self.op_type == 'LinearRegressor':
            self._make_sure_inputs(1)
            coefficients = self._getat('coefficients', None)
            intercepts = self._getat('intercepts', None)
            post_transform = self._getat(
                'post_transform', 'NONE').strip('"\'b')
            targets = self._getat('targets', 1)
            if post_transform != "NONE":
                raise NotImplementedError(  # pragma: no cover
                    "Conversion of operator %r with post_transform=%r "
                    "is not implemented." % (self.op_type, post_transform))
            rows = [
                "coefs = numpy.array(%s, dtype=numpy.float32)."
                "reshape((%d, -1)).T" % (coefficients, targets),
                "%sinter = numpy.array(%s, dtype=numpy.float32)."
                "reshape((-1, %d))" % (self.indent, intercepts, targets),
                f"{self.indent}{outs} = {self.inputs[0]} @ coefs + inter"]
            return "\n".join(rows)

        if self.op_type == 'Normalizer':
            self._make_sure_inputs(1)
            post_transform = self._getat('norm', 'MAX').strip('"\'b')
            if post_transform == 'L2':
                return "%s = %s / (%s ** 2).sum(axis=1) ** 0.5" % (
                    outs, self.inputs[0], self.inputs[0])
            if post_transform == 'L1':
                post_transform = 'sum'
            return "%s = %s / %s.%s(axis=1, keepdims=1)" % (
                outs, self.inputs[0], self.inputs[0], post_transform.lower())

        raise NotImplementedError(  # pragma: no cover
            "Unable to convert operator type %r name=%r (onnxml)." % (
                self.op_type, self.name))

    def _make_numpy_code_others(self):
        outs = ", ".join(self.outputs)

        if self.op_type == 'CDist':
            self._make_sure_inputs(2)
            metric = self._getat('metric', 'euclidean').strip("'b")
            return "%s = scipy_distance.cdist(%s, %s, metric=%r)" % (
                outs, self.inputs[0], self.inputs[1], metric)

        raise NotImplementedError(  # pragma: no cover
            "Unable to convert operator type %r (domain=%r) "
            "name=%r (onnxml)." % (
                self.op_type, self.domain, self.name))


def make_numpy_code(opset, name=None, op_type=None, domain='',
                    inputs=None, outputs=None, attributes=None,
                    used=None, context=None, mark_inits=None,
                    indent="", **unused):
    """
    Converts an ONNX operators into :epkg:`numpy` code.

    :param opset: target opset for the conversion (usually unused)
    :param name: node name
    :param op_type: operator type
    :param domain: domain
    :param inputs: inputs
    :param outputs: outputs
    :param attributes: attributes
    :param used: dictionary `{k: v}`,
        list of nodes taking *k* as input
    :param context: whole context
    :param mark_inits: marks initializer as replaced
    :param indent: indentation of the second line and following
    :return: code as str
    """
    cl = NumpyCode(
        opset=opset, name=name, op_type=op_type, domain=domain,
        inputs=inputs, outputs=outputs, attributes=attributes,
        used=used, context=context, mark_inits=mark_inits,
        indent=indent, **unused)
    return cl.make_numpy_code()



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRun, RefAttrName


def _check_dtype(val):
    a = val.dtype
    if not isinstance(a, numpy.dtype) and a not in {
            numpy.int8, numpy.uint8, numpy.float16, numpy.float32,
            numpy.float64, numpy.int32, numpy.int64, numpy.int16,
            numpy.uint16, numpy.uint32, numpy.bool_, numpy.str_,
            numpy.uint64, bool, str, }:
        raise TypeError(  # pragma: no cover
            f"Type ({a}, {type(a)}) is not a numpy type (operator 'Constant')")


class Constant_9(OpRun):

    atts = {'value': numpy.array([0], dtype=numpy.float32)}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Constant_9.atts,
                       **options)
        self.cst = self.value
        _check_dtype(self.cst)

    def _run(self, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (self.cst, )


class Constant_11(OpRun):

    atts = {'value': numpy.array([0], dtype=numpy.float32),
            'sparse_value': None, }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Constant_11.atts,
                       **options)
        if getattr(self, 'sparse_value', None) is not None:
            self.cst = self.sparse_value
        else:
            self.cst = self.value
        _check_dtype(self.cst)

    def _run(self, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (self.cst, )


class Constant_12(OpRun):

    atts = {'value': numpy.array([0], dtype=numpy.float32),
            'sparse_value': None,
            'value_float': None,
            'value_floats': None,
            'value_int': None,
            'value_ints': None,
            'value_string': None,
            'value_strings': None,
            }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Constant_12.atts,
                       **options)
        if hasattr(self, 'sparse_value') and self.sparse_value is not None:
            self.cst = self.sparse_value
        elif hasattr(self, 'value_float') and self.value_float is not None:
            self.cst = numpy.array([self.value_float], dtype=numpy.float32)
        elif hasattr(self, 'value_floats') and self.value_floats is not None:
            self.cst = self.value_floats.astype(numpy.float32)
        elif hasattr(self, 'value_int') and self.value_int is not None:
            self.cst = self.value_int.astype(numpy.int64)
        elif hasattr(self, 'value_ints') and self.value_ints is not None:
            self.cst = self.value_ints.astype(numpy.int64)
        elif hasattr(self, 'value_string') and self.value_string is not None:
            self.cst = self.value_string
        elif hasattr(self, 'value_strings') and self.value_strings is not None:
            self.cst = self.value_strings
        elif hasattr(self, 'value') and self.value is not None:
            self.cst = self.value
        else:
            raise AttributeError(  # pragma: no cover
                "No constant is defined for operator 'Constant'.")
        if isinstance(self.cst, RefAttrName):
            self.is_linked_attribute = True
        else:
            self.is_linked_attribute = False
            _check_dtype(self.cst)

    def _run(self, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.is_linked_attribute:
            if attributes is None:
                raise RuntimeError(  # pragma: no cover
                    f"Attributes are empty, cannot retrieve value for {self.cst!r}.")
            if self.cst.name not in attributes:
                raise RuntimeError(  # pragma: no cover
                    f"Cannot find attribute {self.cst!r} in {list(attributes)!r}.")
            return (attributes[self.cst.name]['value'], )
        return (self.cst, )


if onnx_opset_version() >= 12:
    Constant = Constant_12
elif onnx_opset_version() >= 11:  # pragma: no cover
    Constant = Constant_11
else:  # pragma: no cover
    Constant = Constant_9



# -*- coding: utf-8 -*-
"""
@file
@brief List of interpreted from scikit-learn model.
"""
import numpy
from .g_sklearn_type_helpers import check_type
from .grammar.exc import Float32InfError
from .grammar.gactions import MLActionCst, MLActionVar, MLActionConcat, MLActionReturn
from .grammar.gactions_num import MLActionAdd, MLActionSign
from .grammar.gactions_tensor import MLActionTensorDot
from .grammar.gmlactions import MLModel


def sklearn_logistic_regression(model, input_names=None, output_names=None, **kwargs):
    """
    Interprets a `logistic regression <http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html>`_
    model into a *grammar* model (semantic graph representation).

    @param      model           *scikit-learn* model
    @param      input_names     name of the input features
    @param      output_names    name of the output predictions
    @param      kwargs          additional parameters (*with_loop*)
    @return                     graph model

    If *input* is None or *output* is None, default values
    will be given to the outputs
    ``['Prediction', 'Score']`` for the outputs.
    If *input_names* is None, it wil be ``'Features'``.

    Additional parameters:
    - *with_loop*: False by default, *True* not implemented.
    - *dtype*: float32 or float64
    """
    if kwargs.get('with_loop', False):
        raise NotImplementedError(  # pragma: no cover
            "Loop version is not implemented.")
    if output_names is None:
        output_names = ['Prediction', 'Score']  # pragma: no cover
    if input_names is None:
        input_names = 'Features'

    from sklearn.linear_model import LogisticRegression
    check_type(model, LogisticRegression)
    if len(model.coef_.shape) > 1 and min(model.coef_.shape) != 1:
        raise NotImplementedError(  # pragma: no cover
            f"Multiclass is not implemented yet: coef_.shape={model.coef_.shape}.")
    dtype = kwargs.get('dtype', numpy.float32)
    coef_ = model.coef_.ravel()
    coef = coef_.astype(dtype)
    bias = dtype(model.intercept_[0])

    for i, c in enumerate(coef):
        if numpy.isinf(c):
            raise Float32InfError(  # pragma: no cover
                f'Unable to convert coefficient {i}: {c}')
    if numpy.isinf(bias):
        raise Float32InfError(  # pragma: no cover
            f'Unable to convert intercept {model.intercept_[0]}')

    gr_coef = MLActionCst(coef)
    gr_var = MLActionVar(coef, input_names)
    gr_bias = MLActionCst(bias)
    gr_dot = MLActionTensorDot(gr_coef, gr_var)
    gr_dist = MLActionAdd(gr_dot, gr_bias)
    gr_sign = MLActionSign(gr_dist)
    gr_conc = MLActionConcat(gr_sign, gr_dist)
    ret = MLActionReturn(gr_conc)
    return MLModel(ret, output_names, name=LogisticRegression.__name__)


def sklearn_linear_regression(model, input_names=None, output_names=None, **kwargs):
    """
    Converts a `linear regression <http://scikit-learn.org/stable/modules/
    generated/sklearn.linear_model.LinearRegression.html>`_
    into a *grammar* model (semantic graph representation).

    @param      model           *scikit-learn* model
    @param      input_names     name of the input features
    @param      output_names    name of the output predictions
    @param      kwargs          additional parameter (*with_loop*)
    @return                     graph model

    If *input* is None or *output* is None, default values
    will be given to the outputs
    ``['Prediction', 'Score']`` for the outputs.
    If *input_names* is None, it wil be ``'Features'``.

    Additional parameters:
    - *with_loop*: False by default, *True* not implemented.
    - *dtype*: float32 or float64
    """
    if kwargs.get('with_loop', False):  # pragma: no cover
        raise NotImplementedError("Loop version is not implemented.")
    if output_names is None:
        output_names = ['Prediction', 'Score']
    if input_names is None:
        input_names = 'Features'

    from sklearn.linear_model import LinearRegression
    check_type(model, LinearRegression)
    if len(model.coef_.shape) > 1 and min(model.coef_.shape) != 1:
        raise NotImplementedError(  # pragma: no cover
            f"MultiOutput is not implemented yet: coef_.shape={model.coef_.shape}.")

    dtype = kwargs.get('dtype', numpy.float32)
    coef_ = model.coef_.ravel()
    coef = coef_.astype(dtype)
    bias = dtype(model.intercept_)

    for i, c in enumerate(coef):
        if numpy.isinf(c):
            raise Float32InfError(  # pragma: no cover
                f'Unable to convert coefficient {i}: {c}')
    if numpy.isinf(bias):
        raise Float32InfError(  # pragma: no cover
            f'Unable to convert intercept {model.intercept_}')

    gr_coef = MLActionCst(coef)
    gr_var = MLActionVar(coef, input_names)
    gr_bias = MLActionCst(bias)
    gr_dot = MLActionTensorDot(gr_coef, gr_var)
    gr_dist = MLActionAdd(gr_dot, gr_bias)
    ret = MLActionReturn(gr_dist)
    return MLModel(ret, output_names, name=LinearRegression.__name__)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRunArg


def _argmax(data, axis=0, keepdims=True):
    result = numpy.argmax(data, axis=axis)
    if keepdims and len(result.shape) < len(data.shape):
        result = numpy.expand_dims(result, axis)
    return result.astype(numpy.int64)


def _argmax_use_numpy_select_last_index(
        data, axis=0, keepdims=True):
    data = numpy.flip(data, axis)
    result = numpy.argmax(data, axis=axis)
    result = data.shape[axis] - result - 1
    if keepdims and len(result.shape) < len(data.shape):
        result = numpy.expand_dims(result, axis)
    return result.astype(numpy.int64)


class _ArgMax(OpRunArg):
    """
    Base class for runtime for operator `ArgMax
    <https://github.com/onnx/onnx/blob/master/docs/
    Operators.md#ArgMax>`_.
    """

    def __init__(self, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRunArg.__init__(self, onnx_node, desc=desc,
                          expected_attributes=expected_attributes,
                          **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (_argmax(data, axis=self.axis, keepdims=self.keepdims), )

    def to_python(self, inputs):
        return ('import numpy\nfrom mlprodict.onnxrt.ops_cpu.op_argmax import _argmax',
                f'return _argmax({inputs[0]}, axis=axis, keepdims=keepdims)')


class ArgMax_11(_ArgMax):

    atts = {'axis': 0, 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        _ArgMax.__init__(self, onnx_node, desc=desc,
                         expected_attributes=ArgMax_11.atts,
                         **options)


class ArgMax_12(_ArgMax):

    atts = {'axis': 0, 'keepdims': 1, 'select_last_index': 0}

    def __init__(self, onnx_node, desc=None, **options):
        _ArgMax.__init__(self, onnx_node, desc=desc,
                         expected_attributes=ArgMax_12.atts,
                         **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.select_last_index == 0:
            return _ArgMax._run(self, data)
        return (_argmax_use_numpy_select_last_index(
            data, axis=self.axis, keepdims=self.keepdims), )

    def to_python(self, inputs):
        lines = [
            "if select_last_index == 0:",
            "    return _argmax({0}, axis=axis, keepdims=keepdims)",
            "return _argmax_use_numpy_select_last_index(",
            "    {0}, axis=axis, keepdims=keepdims)"]
        return ('import numpy\nfrom mlprodict.onnxrt.ops_cpu.op_argmax import _argmax, _argmax_use_numpy_select_last_index',
                "\n".join(lines).format(inputs[0]))


if onnx_opset_version() >= 12:
    ArgMax = ArgMax_12
else:
    ArgMax = ArgMax_11  # pragma: no cover



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# pylint: disable=E1101, C0302

"""
@file
@brief Implements a class able to compute the predictions
from on an :epkg:`ONNX` model.
"""
import hashlib
from collections import Counter
import pprint
from onnx import (
    shape_inference, ModelProto, FunctionProto, GraphProto,
    AttributeProto)
from onnx.helper import (
    make_tensor_value_info, ValueInfoProto, set_model_props,
    make_graph, make_function, make_model, make_node,
    make_operatorsetid, make_attribute, make_value_info)
from .onnx2py_helper import (
    guess_proto_dtype, from_array, get_tensor_shape,
    get_tensor_elem_type)
from .optim import onnx_remove_node_unused
from .onnx_tools import enumerate_onnx_names, enumerate_onnx_nodes
from ..onnx_tools.onnx2py_helper import _var_as_dict, from_array


def enumerate_model_node_outputs(model, add_node=False, order=False):
    """
    Enumerates all the nodes of a model.

    :param model: :epkg:`ONNX` graph
    :param add_node: if False, the function enumerates
        all output names from every node, otherwise, it
        enumerates tuple (output name, node)
    :param order: goes through outputs following the graph order
    :return: enumerator
    """
    if not hasattr(model, "graph"):
        raise TypeError(  # pragma: no cover
            f"Parameter model is not an ONNX model but {type(model)}")
    if order:
        edges = []
        order = {}
        node_names = {}
        for inp in model.graph.input:
            order[0, inp.name] = 0
        for node in model.graph.node:
            order[1, node.name] = 0
            for i in node.input:
                edges.append(('in', i, node.name))
            for o in node.output:
                edges.append(('out', o, node.name))
                node_names[o] = node
                order[0, o] = 0

        modif = 1
        n_iter = 0
        while modif > 0 and n_iter <= len(model.graph.node):
            modif = 0
            n_iter += 1
            for kind, data_name, node_name in edges:
                if kind == 'in':
                    if (0, data_name) not in order:
                        continue
                    if order[0, data_name] + 1 > order[1, node_name]:
                        modif += 1
                        order[1, node_name] = order[0, data_name] + 1
                else:
                    if order[1, node_name] + 1 > order[0, data_name]:
                        modif += 1
                        order[0, data_name] = order[1, node_name] + 1

        orders = [(v, k) for k, v in order.items()]
        orders.sort()

        for _, k in orders:
            if k[0] == 1:
                continue
            out = k[1]
            if out not in node_names:
                continue
            yield (out, node_names[out]) if add_node else out
    else:
        for node in model.graph.node:
            for out in node.output:
                yield (out, node) if add_node else out


def get_opsets(model, include_functions=True, exc=True):
    """
    Enumerates all opsets used in a model.

    :param model: :epkg:`ModelProto` or :epkg:`FunctionProto`
    :param include_functions: include opsets used in functions
    :param exc: raise an exception if conflicts are detected
    :return: dictionary
    """
    if isinstance(model, ModelProto):
        res = {}
        for op in model.opset_import:
            if exc and op.domain in res:
                raise ValueError(  # pragma: no cover
                    f"Domain {op.domain!r} appears multiple times.")
            res[op.domain] = op.version
        if include_functions:
            for f in model.functions:
                ops = get_opsets(f, exc=exc)
                for k, v in ops.items():
                    if k in res:
                        if res[k] != v:
                            if exc:
                                raise ValueError(  # pragma: no cover
                                    "Domain %r has different version in "
                                    "main graph (%d) and function %r "
                                    "(%d)." % (k, res[k], f.name, v))
                            res[k] = max(res[k], v)
                    else:
                        res[k] = v
        return res

    res = {}
    for op in model.opset_import:
        if exc and op.domain in res:
            raise ValueError(  # pragma: no cover
                f"Domain {op.domain!r} appears multiple times.")
        res[op.domain] = op.version
    return res


def get_hidden_inputs(nodes):
    """
    Returns the list of hidden inputs used by subgraphs.

    :param nodes: list of nodes
    :return: list of names
    """
    inputs = set()
    outputs = set()
    for node in nodes:
        inputs |= set(node.input)
        outputs |= set(node.output)
        for att in node.attribute:
            if (att.type != AttributeProto.GRAPH or  # pylint: disable=E1101
                    not hasattr(att, 'g') or att.g is None):
                continue
            hidden = get_hidden_inputs(att.g.node)
            inits = set(att.g.initializer)
            inputs |= hidden - (inits & hidden)
    return inputs - (outputs & inputs)


def select_model_inputs_outputs(model, outputs=None, inputs=None,
                                infer_shapes=False, overwrite=None,
                                remove_unused=True,
                                verbose=0, fLOG=None):
    """
    Takes a model and changes its outputs.

    :param model: :epkg:`ONNX` model
    :param inputs: new inputs, same ones if None
    :param outputs: new outputs, same ones if None
    :param infer_shapes: infer inputs and outputs shapes
    :param overwrite: overwrite type and shapes for
        inputs or outputs, *overwrite* is a
        dictionary `{'name': (numpy dtype, shape)}`
    :param remove_unused: remove unused nodes from the graph
    :param verbose: display information while converting
    :param fLOG: logging function
    :return: modified model

    The function removes unneeded nodes.

    .. exref::
        :title: Change ONNX model inputs

        The following exampels shows how to change the inputs of model
        to bypass the first nodes. Shape inferences fails to determine
        the new inputs type. They need to be overwritten.
        `verbose=1, fLOG=print` shows the number of deleted nodes.

        ::

            import onnx
            from mlprodict.onnx_tools.onnx_manipulations import select_model_inputs_outputs

            onx = onnx.load(path)
            onx2 = select_model_inputs_outputs(
                onx, inputs=["SentenceTokenizer/SentencepieceTokenizeOp:0",
                             "SentenceTokenizer/SentencepieceTokenizeOp:1"],
                infer_shapes=True, verbose=1, fLOG=print,
                overwrite={'SentenceTokenizer/SentencepieceTokenizeOp:0': (numpy.int32, None),
                           'SentenceTokenizer/SentencepieceTokenizeOp:1': (numpy.int64, None)})
            onnx.save(onx2, path2)
    """
    if inputs is not None and not isinstance(inputs, list):
        inputs = [inputs]
    if outputs is not None and not isinstance(outputs, list):
        outputs = [outputs]
    if inputs is None:
        inputs = [i.name for i in model.graph.input]
    if outputs is None:
        outputs = [o.name for o in model.graph.output]

    mark_var = {}
    for out in enumerate_model_node_outputs(model):
        mark_var[out] = 0
    for inp in inputs:
        mark_var[inp] = 0
    for out in outputs:
        if out not in mark_var:
            raise ValueError(  # pragma: no cover
                f"Output '{out}' not found in model.")
        mark_var[out] = 1

    nodes = model.graph.node[::-1]
    mark_op = {}
    for node in nodes:
        mark_op[id(node)] = 0

    # We mark all the nodes we need to keep.
    nb = 1
    while nb > 0:
        nb = 0
        for node in nodes:
            if mark_op[id(node)] == 1:
                continue
            mod = False
            for out in node.output:
                if mark_var[out] == 1:
                    mark_op[id(node)] = 1
                    mod = True
                    break
            if not mod:
                continue

            hidden = get_hidden_inputs([node])
            node_inputs = list(node.input) + list(hidden)

            nb += 1
            for inp in node_inputs:
                if inp in inputs:
                    continue
                if mark_var.get(inp, 0) == 1:
                    continue
                mark_var[inp] = 1
                nb += 1

    # All nodes verifies mark_op[node.name] == 1
    keep_nodes = [node for node in nodes[::-1] if mark_op[id(node)] == 1]

    if verbose > 1 and fLOG is not None:  # pragma: no cover
        for node in nodes:
            s = "+" if mark_op[id(node)] == 1 else "-"
            fLOG("[select_model_inputs_outputs] %s %s (%s) -> %s [%s]" % (
                s, node.op_type, ", ".join(node.input),
                ', '.join(node.output), node.name))

    known_shapes = {}
    if infer_shapes:
        shapes = shape_inference.infer_shapes(model)
        for shape in shapes.graph.value_info:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.input:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.output:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
    else:
        for shape in model.graph.input:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in model.graph.output:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type

    var_in = []
    for name in inputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = guess_proto_dtype(dtype)
            value_info = make_tensor_value_info(
                name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = ValueInfoProto()
                value_info.name = name
            else:
                shape = get_tensor_shape(known_shapes[name])
                value_info = make_tensor_value_info(
                    name, proto_dtype, shape)
        else:
            value_info = ValueInfoProto()
            value_info.name = name
        var_in.append(value_info)

    var_out = []
    for name in outputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = guess_proto_dtype(dtype)
            value_info = make_tensor_value_info(
                name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = ValueInfoProto()
                value_info.name = name
            else:
                shape = get_tensor_shape(known_shapes[name])
                value_info = make_tensor_value_info(
                    name, proto_dtype, shape)
        else:
            value_info = ValueInfoProto()
            value_info.name = name
        var_out.append(value_info)

    if verbose > 0 and fLOG is not None:  # pragma: no cover
        fLOG("[select_model_inputs_outputs] nodes %r --> %r" % (
            len(model.graph.node), len(keep_nodes)))
        fLOG("[select_model_inputs_outputs] inputs: %r" %
             [_.name for _ in var_in])
        fLOG("[select_model_inputs_outputs] inputs: %r" %
             [_.name for _ in var_out])

    graph = make_graph(keep_nodes, model.graph.name, var_in,
                       var_out, model.graph.initializer,
                       sparse_initializer=model.graph.sparse_initializer)
    onnx_model = make_model(graph, functions=model.functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version

    # remove unused nodes
    if remove_unused:
        onnx_model = onnx_remove_node_unused(onnx_model, recursive=False)

    return onnx_model


def change_input_type(onx, changes):
    """
    Changes the type of an input.

    :param onx: ONNX model
    :param changes: dictionary '{ name: new proto element type }`
    :return: new onx
    """
    new_inputs = []
    for inp in onx.graph.input:
        if inp.name not in changes:
            new_inputs.append(inp)
            continue
        value_info = make_tensor_value_info(
            inp.name, changes[inp.name], None)
        new_inputs.append(value_info)

    # final
    graph = make_graph(list(onx.graph.node),
                       onx.graph.name, new_inputs,
                       list(onx.graph.output),
                       onx.graph.initializer,
                       sparse_initializer=onx.graph.sparse_initializer)
    onnx_model = make_model(graph, functions=onx.functions)
    onnx_model.ir_version = onx.ir_version
    onnx_model.producer_name = onx.producer_name
    onnx_model.producer_version = onx.producer_version
    onnx_model.domain = onx.domain
    onnx_model.model_version = onx.model_version
    onnx_model.doc_string = onx.doc_string
    if len(onx.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in onx.metadata_props}
        set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in onx.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return onnx_model


def _change_subgraph_io_type_shape_list(io_list, type_changes, shape_changes):
    ms = False
    new_inputs = []
    for inp in io_list:
        m = False
        if isinstance(shape_changes, dict):
            if inp.name in shape_changes:
                shape = shape_changes[inp.name]
                m = True
            else:
                shape = get_tensor_shape(inp)
        else:
            shape = shape_changes(inp)
            m = True

        if isinstance(type_changes, dict):
            if inp.name in type_changes:
                ntype = type_changes[inp.name]
                m = True
            else:
                ntype = get_tensor_elem_type(inp)
        else:
            ntype = type_changes(inp)
            m = True

        if m:
            ms = True
            value_info = make_tensor_value_info(inp.name, ntype, shape)
            new_inputs.append(value_info)
        else:
            new_inputs.append(inp)
    return new_inputs if ms else None


def change_subgraph_io_type_shape(onx, type_changes=None, shape_changes=None,
                                  recursive=True):
    """
    Changes the type of an input or an output of a subgraph.

    :param onx: ModelProto, GraphProto
    :param type_changes: dictionary '{ name: new proto element type }`
        or function `f(input) -> type`
    :param shape_changes: dictionary '{ name: new shape }`
        or function `f(input) -> shape`
    :param recursive: True
    :return: new onx
    """
    if isinstance(onx, ModelProto):
        graph = change_subgraph_io_type_shape(
            onx.graph, type_changes, shape_changes, recursive)
        onnx_model = make_model(graph, functions=onx.functions)
        onnx_model.ir_version = onx.ir_version
        onnx_model.producer_name = onx.producer_name
        onnx_model.producer_version = onx.producer_version
        onnx_model.domain = onx.domain
        onnx_model.model_version = onx.model_version
        onnx_model.doc_string = onx.doc_string
        if len(onx.metadata_props) > 0:  # pragma: no cover
            values = {p.key: p.value for p in onx.metadata_props}
            set_model_props(onnx_model, values)

        del onnx_model.opset_import[:]  # pylint: disable=E1101
        for oimp in onx.opset_import:
            op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
            op_set.domain = oimp.domain
            op_set.version = oimp.version
        return onnx_model

    graph = onx
    new_inputs = _change_subgraph_io_type_shape_list(
        graph.input, type_changes or {}, shape_changes or {})
    if new_inputs is None:
        new_inputs = graph.input

    new_outputs = _change_subgraph_io_type_shape_list(
        graph.output, type_changes or {}, shape_changes or {})
    if new_outputs is None:
        new_outputs = graph.output

    # recursive
    if recursive:
        new_nodes = []
        for node in graph.node:
            modified = False
            atts = []
            for att in node.attribute:
                if (att.type == AttributeProto.GRAPH and
                        hasattr(att, 'g') and att.g is not None):
                    modified = True
                    g = change_subgraph_io_type_shape(
                        att.g, type_changes, shape_changes,
                        recursive=recursive)
                    att = make_attribute(att.name, g)
                atts.append(att)
            if modified:
                node = make_node(node.op_type, node.input, node.output)
                node.attribute.extend(atts)
            new_nodes.append(node)
    else:
        new_nodes = list(graph.node)

    # final
    graph = make_graph(new_nodes, graph.name, new_inputs, new_outputs,
                       graph.initializer,
                       sparse_initializer=graph.sparse_initializer)
    return graph


def overwrite_opset(model, new_opset):
    """
    Overwrites the main opset in an ONNX file.
    Does not change any node definition.

    :param model: ONNX model
    :param new_opset: new opset
    :return: ONNX model
    """
    graph = make_graph(
        model.graph.node, model.graph.name, model.graph.input,
        model.graph.output, model.graph.initializer,
        sparse_initializer=model.graph.sparse_initializer)
    onnx_model = make_model(graph, functions=model.functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = new_opset if oimp.domain == '' else oimp.version
    return onnx_model


def hash_onnx_object(obj, max_size):
    """
    Hashes the content of an object.
    It uses module :mod:`hashlib`.

    :param obj: onnx graph (it must have a method `SerializeToString`)
    :param max_size: size of the hash
    :return: hash
    """
    m = hashlib.sha256()
    if hasattr(obj, 'op_type'):
        # An operator.
        m.update(obj.op_type.encode('ascii'))
        m.update(str(len(obj.input)).encode('ascii'))
        m.update(str(len(obj.output)).encode('ascii'))
        if hasattr(obj, 'attribute'):
            for att in obj.attribute:
                m.update(att.name.encode('ascii'))
                m.update(att.SerializeToString())
    else:
        # An initializer.
        name = obj.name
        docf = obj.doc_string
        obj.name = ''
        obj.doc_string = ''
        try:
            m.update(obj.SerializeToString())
        except AttributeError as e:  # pragma: no cover
            raise RuntimeError(
                f"Unable to hash object type {type(obj)!r}, value={obj!r}.") from e
        finally:
            obj.name = name
            obj.doc_string = docf

    content = m.hexdigest()
    if len(content) > max_size:
        content = content[:max_size]
    return content.upper()


def onnx_rename_names(model, strategy='simple', recursive=True,
                      verbose=0, fLOG=print,
                      counts=None, replace=None, taken=None):
    """
    Renames all names except the inputs and outputs.

    :param model: onnx model
    :param strategy: two strategies are implemented, see below
    :param recursive: walk through subgraphs
    :param verbose: verbose, if positive, reports on all changed names
    :param fLOG: logging function
    :param counts: used for recursion
    :param replace: used for recursion, it can be also used to
        to fix some replacements
    :param taken: used for recursion
    :return: onnx model (the model is modified in place)

    Strategies:

    * `'simple'`: use a letter `n` for node, `r`, `i` for initializer,
        this letter is followed by a number
    * `'type'`: the name depends on the node type and content,
        the hash is kept as small as possible
    """
    counts = counts or {'init': 0, 'node': 0, 'result': 0}
    replace = replace or {}
    taken = taken or set()
    graph = model.graph if hasattr(model, 'graph') else model

    for obj in graph.input:
        replace[obj.name] = obj.name
    for obj in graph.output:
        replace[obj.name] = obj.name

    def _check_name_simple(prefix):
        if prefix not in replace:
            return prefix
        c = 1
        final = "%s_%d" % (prefix, c)
        while final in taken:
            c += 1
            final = "%s_%d" % (prefix, c)
        taken.add(final)
        return final

    def _check_name_type(obj, prefix):
        c = 2
        hash = hash_onnx_object(obj, c)
        final = f"{prefix}_{hash}"
        while final in taken:
            c += 2
            hash = hash_onnx_object(obj, c)
            final = f"{prefix}_{hash}"
        taken.add(final)
        return final

    def get_name_init(init):
        if init.name in replace:
            return replace[init.name]
        if strategy == 'simple':
            name = _check_name_simple('i%d' % counts['init'])
            counts['init'] += 1
            replace[init.name] = name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] init: {init.name!r} -> {name!r}')
            return name
        if strategy == 'type':
            name = _check_name_type(init, 'i')
            counts['init'] += 1
            replace[init.name] = name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] init: {init.name!r} -> {name!r}')
            return name
        raise ValueError(  # pragma: no cover
            f"Unknown strategy {strategy!r}.")

    def get_name_node(node):
        node_name = 'node_%s_%d' % (node.name, id(node))
        if node_name in replace:
            return replace[node_name]
        if strategy == 'simple':
            name = _check_name_simple('n%d' % counts['node'])
            counts['node'] += 1
            replace[node_name] = name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] node: {node_name!r} -> {name!r}')
            return name
        if strategy == 'type':
            name = _check_name_type(node, 'n')
            counts['node'] += 1
            replace[node_name] = name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] node: {node_name!r} -> {name!r}')
            return name
        raise ValueError(  # pragma: no cover
            f"Unknown strategy {strategy!r}.")

    def get_name_result(node, i, name, suffix):
        if name in replace:
            return replace[name]
        if strategy == 'simple':
            new_name = _check_name_simple('r%d' % counts['result'])
            counts['result'] += 1
            replace[name] = new_name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] result: {name!r} -> {new_name!r}')
            return new_name
        if strategy == 'type':
            new_name = _check_name_type(node, 'r%s%d' % (suffix, i))
            counts['result'] += 1
            replace[name] = new_name
            if verbose > 0 and fLOG is not None:
                fLOG(f'[onnx_rename_names] result: {name!r} -> {new_name!r}')
            return new_name
        raise ValueError(  # pragma: no cover
            f"Unknown strategy {strategy!r}.")

    def get_name_input(node, i):
        return get_name_result(node, i, node.input[i], 'i')

    def get_name_output(node, i):
        return get_name_result(node, i, node.output[i], 'o')

    for init in graph.initializer:
        init.name = get_name_init(init)
    for init in graph.sparse_initializer:
        init.name = get_name_init(init)

    for node in graph.node:
        node.name = get_name_node(node)
        for i in range(len(node.input)):  # pylint: disable=C0200
            node.input[i] = get_name_input(node, i)
        for i in range(len(node.output)):  # pylint: disable=C0200
            node.output[i] = get_name_output(node, i)
        if not recursive or node.op_type not in {'Scan', 'If', 'Loop'}:
            continue
        # recursion
        for att in node.attribute:
            if att.name not in {'if_branch', 'else_branch', 'body'}:
                continue
            onnx_rename_names(
                att.g, strategy=strategy, fLOG=fLOG, verbose=verbose,
                counts=counts, replace=replace, taken=taken)

    return model


def onnx_rename_inputs_outputs(onx, rename):
    """
    Renames input or outputs names.

    :param onx: GraphProto, ModelProto
    :param rename: dictionary `{old_name: new_name}`
    :return: new onx
    """
    if isinstance(onx, ModelProto):
        graph = onnx_rename_inputs_outputs(onx.graph, rename)
        onnx_model = make_model(graph, functions=onx.functions)
        onnx_model.ir_version = onx.ir_version
        onnx_model.producer_name = onx.producer_name
        onnx_model.producer_version = onx.producer_version
        onnx_model.domain = onx.domain
        onnx_model.model_version = onx.model_version
        onnx_model.doc_string = onx.doc_string
        if len(onx.metadata_props) > 0:  # pragma: no cover
            values = {p.key: p.value for p in onx.metadata_props}
            set_model_props(onnx_model, values)

        del onnx_model.opset_import[:]  # pylint: disable=E1101
        for oimp in onx.opset_import:
            op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
            op_set.domain = oimp.domain
            op_set.version = oimp.version
        return onnx_model

    graph = onx
    new_inputs = []
    for inp in graph.input:
        if inp.name not in rename:
            new_inputs.append(inp)
            continue
        value_info = make_tensor_value_info(
            rename[inp.name], get_tensor_elem_type(inp), get_tensor_shape(inp))
        new_inputs.append(value_info)

    new_outputs = []
    for inp in graph.output:
        if inp.name not in rename:
            new_outputs.append(inp)
            continue
        value_info = make_tensor_value_info(
            rename[inp.name], get_tensor_elem_type(inp), get_tensor_shape(inp))
        new_outputs.append(value_info)

    new_inits = []
    for init in graph.initializer:
        if init.name in rename:
            init.name = rename[init.name]
        new_inits.append(init)

    new_sparse_inits = []
    for init in graph.sparse_initializer:
        if init.name in rename:
            init.name = rename[init.name]
        new_sparse_inits.append(init)

    new_nodes = []
    for node in graph.node:
        modified = False
        atts = []
        for att in node.attribute:
            if (att.type == AttributeProto.GRAPH and
                    hasattr(att, 'g') and att.g is not None):
                modified = True
                g = onnx_rename_inputs_outputs(att.g, rename)
                att = make_attribute(att.name, g)
            atts.append(att)
        if modified:
            node = make_node(node.op_type, node.input, node.output)
            node.attribute.extend(atts)

        inp = [rename.get(i, i) for i in node.input]
        out = [rename.get(i, i) for i in node.output]
        if inp == list(node.input) and out == list(node.output):
            new_nodes.append(node)
            continue

        node = make_node(node.op_type, inp, out, domain=node.domain,
                         name=node.name)
        node.attribute.extend(atts)
        new_nodes.append(node)

    # final
    graph = make_graph(new_nodes, graph.name, new_inputs, new_outputs,
                       new_inits, sparse_initializer=new_sparse_inits)
    return graph


def onnx_replace_functions(model, replace):
    """
    Replaces some of the function in model.

    :param model: *ModelProto*
    :param replace: dictionary `{ (domain, name): FunctionProto }`
    :return: new model
    """
    if not isinstance(model, ModelProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(model)!r}.")
    new_functions = []
    modified = False
    for fct in model.functions:
        key = fct.domain, fct.name
        if key in replace:
            modified = True
            f = replace[key]
            if not isinstance(f, FunctionProto):
                raise TypeError(  # pragma: no cover
                    f"Unexpected type {type(f)!r} for function {key!r} in replace.")
            if len(f.input) != len(fct.input):
                raise ValueError(  # pragma: no cover
                    f"Input mismatches {f.input!r} != {fct.input!r} (expected).")
            if len(f.output) != len(fct.output):
                raise ValueError(  # pragma: no cover
                    f"Output mismatches {f.output!r} != {fct.output!r} (expected).")
            new_functions.append(f)
        else:
            new_functions.append(fct)
    if not modified:
        return model
    opsets = [make_operatorsetid(op.domain, op.version)
              for op in model.opset_import]
    onnx_model = make_model(
        model.graph, opset_imports=opsets, functions=new_functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        set_model_props(onnx_model, values)
    return onnx_model


def insert_results_into_onnx(model, results, as_parameter=True, suffix='_DBG',
                             param_name=None, node_type='DEBUG',
                             domain='DEBUG', domain_opset=1):
    """
    Inserts results into an ONNX graph to produce an extended
    ONNX graph. It can be saved and looked into with a tool such as
    :epkg:`netron`.

    :param model: ONNX graph
    :param results: results to be added in a dictionary
    :param as_parameter: add new nodes with results as one parameter
        (True) or as initializer (False)
    :param suffix: suffix to add to new results
    :param param_name: name of the parameter to add
        (by default the result name), it can be a function
        `param_name(reult_name) -> parameter_name`
    :param node_type: type of the new node
    :param domain: domain the new node
    :param domain_opset: opset for *domain*
    :return: new ONNX graph

    See method :meth:`OnnxInference.run2onnx
    <mlprodict.onnxrt.onnx_inference.OnnxInference.run2onnx>`
    to see a graph this function produces.

    .. image:: debug.png

    .. versionadded:: 0.7
    """
    inputs = list(model.graph.input)
    outputs = list(model.graph.output)
    inits = list(model.graph.initializer)
    inits_sparse = list(model.graph.sparse_initializer)
    nodes = {id(n): n for n in model.graph.node}
    order = {id(n): i for i, n in enumerate(model.graph.node)}
    nodes_copy = {}

    names_init = (set(init.name for init in inits) |
                  set(init.name for init in inits_sparse))
    names_input = set(init.name for init in inputs)
    names_output = {}
    for node in nodes.values():
        for i, o in enumerate(node.output):
            names_output[o] = (i, node)

    for k, v in results.items():
        if k in names_init:
            # initializer are not inserted again
            continue
        if k in names_input:
            # inputs are added as
            raise NotImplementedError(
                f"Unable to add debug information on input {k!r}.")

        if k not in names_output:
            raise RuntimeError(
                "Unable to find result %r in the ONNX graph. Available="
                "[%s]." % (k, ", ".join(sorted(names_output))))

        index, node = names_output[k]
        new_name = k + suffix

        if id(node) not in nodes_copy:
            new_node = make_node(
                node.op_type, list(node.input), list(node.output),
                domain=node.domain if node.domain else None,
                name=node.name + suffix)
            new_node.attribute.extend(node.attribute)  # pylint: disable=E1101
            nodes_copy[id(node)] = new_node
            order[id(new_node)] = order[id(node)]
        new_node = nodes_copy[id(node)]
        new_node.output[index] = new_name

        if as_parameter:
            pname = k if param_name is None else param_name(k)
            atts = {pname: from_array(v, name=pname)}
            inserted_node = make_node(
                node_type, [new_name], [k], domain=domain,
                **atts)
        else:
            pname = k if param_name is None else param_name(k)
            pname += suffix + 'i'
            inserted_node = make_node(
                node_type, [new_name, pname], [k], domain=domain)
            inits.append(from_array(v, name=pname))

        order[id(inserted_node)] = order[id(node)] + 1. / (index + 2)
        nodes[id(inserted_node)] = inserted_node

    new_nodes = [(order[id(n)], n)
                 for n in nodes.values() if id(n) not in nodes_copy]
    new_nodes.extend((order[id(n)], n) for n in nodes_copy.values())
    new_nodes = [n[1] for n in sorted(new_nodes)]

    graph = make_graph(new_nodes, model.graph.name, inputs, outputs,
                       inits, sparse_initializer=inits_sparse)
    onnx_model = make_model(graph, functions=model.functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
    op_set.domain = domain
    op_set.version = domain_opset
    return onnx_model


def onnx_model_to_function(onx, name=None, domain="custom",
                           opset_imports=None, doc_string=None,
                           inputs2par=None):
    """
    Converts an ONNX model into a function. The returned function
    has no attribute.

    :param onx: onnx model
    :param name: function name
    :param domain: function domain
    :param opset_imports: opset to import as a dictionary
        `{domain: version}`
    :param doc_string: doc string
    :param inputs2par: dictionary to move some inputs as attributes
        `{ name: None or default value }`
    :return: function, other functions

    .. warning::
        :epkg:`FunctionProto` does not support default values yet.
        They are ignored.
    """
    if isinstance(onx, ModelProto):
        if opset_imports is None:
            domains = {}
            for op in onx.opset_import:
                domains[op.domain] = op.version
            opset_imports = domains
        if doc_string is None:
            doc_string = onx.doc_string
        fp, lf = onnx_model_to_function(
            onx.graph, name=name, domain=domain,
            opset_imports=opset_imports, doc_string=doc_string,
            inputs2par=inputs2par)
        return fp, lf + list(onx.functions)

    if not isinstance(onx, GraphProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(onx)!r} for onx.")

    if name is None:
        name = onx.name

    inputs = []
    outputs = [o.name for o in onx.output]
    attributes = []
    nodes = []
    if inputs2par is None:
        inputs.extend(i.name for i in onx.input)
    else:
        for i in onx.input:
            if i.name not in inputs2par:
                inputs.append(i.name)
                continue
            attributes.append(i.name)

    if len(onx.initializer) > 0 or len(onx.sparse_initializer) > 0:
        # Needs to convert every initializer into Constant.
        csts = []
        for init in onx.initializer:
            v = _var_as_dict(init)
            value = from_array(v['value'])
            n = make_node('Constant', [], [init.name], value=value)
            csts.append(n)
        for init in onx.sparse_initializer:
            v = _var_as_dict(init)
            value = from_array(v['sparse_value'])
            n = make_node('Constant', [], [init.name], sparse_value=value)
            csts.append(n)
        nodes.extend(csts)

    nodes.extend(onx.node)

    if isinstance(opset_imports, dict):
        ops = [make_operatorsetid(k, v) for k, v in opset_imports.items()]
        opset_imports = ops
    return make_function(
        domain, name, inputs, outputs, nodes,
        opset_imports=opset_imports, doc_string=doc_string or '',
        attributes=attributes), []


def _onnx_function_to_model_convert_io(ens, type_info, shape_fct):
    typed_io = []
    for name in ens:
        if isinstance(type_info, dict):
            res = type_info[name]
        elif callable(type_info):
            res = type_info(name)
        else:
            raise TypeError(  # pragma: no cover
                "type_info is not a callable or a dictionary, "
                "unable to guess type for name=%r with "
                "type(type_info)=%r." % (name, type(type_info)))
        if isinstance(res, int):
            proto_dtype = res
        else:
            proto_dtype = guess_proto_dtype(res)
        value_info = make_tensor_value_info(
            name, proto_dtype, shape_fct(name, proto_dtype))
        typed_io.append(value_info)
    return typed_io


def onnx_function_to_model(onx, functions=None, type_info=None,
                           as_function=False, shape_fct=None):
    """
    Converts an ONNX FunctionProto into a ModelProto.
    The function does not handle attributes yet.

    :param onx: onnx function
    :param functions: additional functions to append to the model
    :param type_info: dictionary or callable which returns the type of
        inputs or outputs if it cannot be guessed
    :param as_function: if True, the function stays as a function and a single node
        is created to call that function
    :param shape_fct: function to specify the shapes,
        signature: `shape_fct(name, proto_type) -> list`
    :return: function
    """
    if not isinstance(onx, FunctionProto):
        raise TypeError(  # pragma: no cover
            f"onx must be a FunctionProto not {type(onx)!r}.")
    if len(onx.attribute) > 0:
        raise NotImplementedError(
            "The function has attributes, it is not implemented yet.")

    if isinstance(functions, list):
        added_functions = functions.copy()
    elif isinstance(functions, dict):
        added_functions = list(functions.values())
    elif functions is None:
        added_functions = []
    else:
        raise TypeError(  # pragma: no cover
            f"Unexpected type for functions {type(functions)!r}.")

    if shape_fct is None:
        shape_fct = lambda name, dtype: None

    inputs = _onnx_function_to_model_convert_io(
        onx.input, type_info, shape_fct=shape_fct)
    outputs = _onnx_function_to_model_convert_io(
        onx.output, type_info, shape_fct=shape_fct)
    if as_function:
        nodes = [make_node(onx.name,
                           [i.name for i in inputs],
                           [o.name for o in outputs],
                           domain=onx.domain)]
        added_functions.append(onx)
        opsets = [make_operatorsetid(onx.domain, 1)]
    else:
        nodes = list(onx.node)
        opsets = [make_operatorsetid(op.domain, op.version)
                  for op in onx.opset_import]
    graph = make_graph(nodes, onx.name, inputs, outputs,
                       [], doc_string=onx.doc_string)
    model = make_model(graph, functions=added_functions,
                       opset_imports=opsets,
                       doc_string=onx.doc_string,
                       model_version=1,
                       domain=onx.domain)
    return model


def _get_new_name(prefix, name, existing_names):
    opt = f"{prefix}_{name}_0"
    i = 0
    while opt in existing_names:
        i += 1
        opt = "%s_%s_%d" % (prefix, name, i)
    existing_names.add(opt)
    return opt


def onnx_subgraphs_level(obj):
    """
    Returns the depth of the graph.

    :param obj: onnx object
    :return: integer
    """
    if isinstance(obj, ModelProto):
        return onnx_subgraphs_level(obj.graph)
    best = 0
    for node in obj.node:
        for att in node.attribute:
            if (att.type == AttributeProto.GRAPH and
                    hasattr(att, 'g') and att.g is not None):
                m = onnx_subgraphs_level(att.g)
                if m > best:
                    best = m
    return best + 1


class _inline_mapping(dict):
    """
    Overwrites class dictionary to debug more easily.

    :param verbose: verbosity
    :param fLOG: logging function
    :param level: sub graph level
    """

    def __init__(self, verbose, fLOG, level):
        dict.__init__(self)
        self._verbose = verbose
        self._fLOG = fLOG
        self._level = level

    def __setitem__(self, key, value):
        "Adds a value."
        if self._verbose > 3:
            self._fLOG("[_inline_mapping-dict-addkv] %s + %r: %r" %
                       ("  " * self._level, key, value))
        if key in self:
            raise RuntimeError(  # pragma: no cover
                "Key %r was already added (with value %r, new one is %r)."
                "" % (key, self[key], value))
        dict.__setitem__(self, key, value)

    def update(self, d):
        "Updates many values."
        for k, v in d.items():
            self[k] = v

    def copy(self):
        "Returns a copy."
        m = _inline_mapping(self._verbose, self._fLOG, self._level)
        for k, v in self.items():
            m[k] = v
        return m

    def remove(self, o):
        "Removes one element."
        if o not in self:
            raise KeyError(  # pragma: no cover
                f"Cannot remove a key {o!r}.")
        self.pop(o)


def _onnx_inline_function_graph(graph, protos, existing_names, mapping,
                                verbose, fLOG, rename, level):
    if len(graph.node) == 0:
        # Outputs have still to be renamed.
        graph0 = graph
        if verbose > 1:
            fLOG(  # pragma: no cover
                "[onnx_inline_function-graph] %s visit0 graph=%d rename=%r "
                "len(mapping)=%d begin" % (
                    "  " * level, id(graph), rename, len(mapping)))
        if rename:
            modified_nodes = []
            mapping = mapping.copy()
            for i in graph.input:
                mapping[i.name] = i.name
            for i in graph.initializer:
                mapping[i.name] = i.name
            for i in graph.sparse_initializer:
                mapping[i.name] = i.name
            outputs = []
            for o in graph.output:
                no = make_value_info(mapping[o.name], o.type)
                if no.name != o.name:
                    modified_nodes.append(o)
                    outputs.append(no)
                else:
                    outputs.append(o)
            if len(modified_nodes) > 0:
                graph = make_graph(
                    [], graph.name, graph.input, outputs,
                    graph.initializer, doc_string=graph.doc_string,
                    sparse_initializer=list(graph.sparse_initializer))
        else:
            modified_nodes = []

        if verbose > 1:
            fLOG(  # pragma: no cover
                "[onnx_inline_function-graph] %s visit graph=%d end "
                "changed=%r len(modified_nodes)=%d" % (
                    "  " * level, id(graph0), id(graph0) != id(graph),
                    len(modified_nodes)))

        return graph, modified_nodes

    graph0 = graph
    mapping = mapping.copy()
    init = list(graph.initializer)
    init_sparse = list(graph.sparse_initializer)
    inputs = list(graph.input)
    modified_nodes = []
    outputs = list(graph.output)

    if verbose > 1:
        fLOG("[onnx_inline_function-graph] %s >visit graph=%d rename=%r "
             "len(mapping)=%d begin" % (
                 "  " * level, id(graph), rename, len(mapping)))

    output_names = [o.name for o in outputs]
    for i in init:
        mapping[i.name] = i.name
    for i in init_sparse:
        mapping[i.name] = i.name
    for i in inputs:
        mapping[i.name] = i.name

    # first step, replace names
    nodes = []
    for node in graph.node:
        mod = 0
        inp = []
        for i in node.input:
            if i in mapping:
                inp.append(mapping[i])
                if mapping[i] != i:
                    mod += 1
            else:
                raise RuntimeError(  # pragma: no cover
                    "Cannot find input %r in %s for node (level=%d)\n%r." % (
                        i, pprint.pformat(mapping), level, node))
        out = []
        for o in node.output:
            new_o = o
            if rename:
                if o not in output_names:
                    new_o = _get_new_name('_inl', o, existing_names)
                if o in mapping:
                    # See below.
                    mapping.remove(o)
            elif o in mapping:
                # That means the main contains a result node but is overwritten by
                # the subgraph. The local variable cannot be reached anymore,
                # we remove it.
                mapping.remove(o)
                if o in node.input:
                    new_o = _get_new_name('_inl', o, existing_names)
                if verbose > 3:
                    fLOG(
                        "[onnx_inline_function-renam] %s node %r(%r): %r -> %r "
                        "overwrite result (%r -> %r)." % (
                            "  " * level, node.op_type, node.name, node.input,
                            node.output, o, new_o))
            out.append(new_o)
            mapping[o] = new_o
            if o != new_o:
                mapping[new_o] = new_o
                mod += 1

        if verbose > 3:
            fLOG("[onnx_inline_function-renam] %s rep node %r(%r): %r -> %r" % (
                "  " * level, node.op_type, node.name, node.input, node.output))
        new_node = make_node(node.op_type, inp, out, domain=node.domain,
                             name=_get_new_name('_inln', node.name, existing_names))
        for att in node.attribute:
            if (att.type == AttributeProto.GRAPH and
                    hasattr(att, 'g') and att.g is not None):
                g, m = _onnx_inline_function_graph(
                    att.g, protos, existing_names=existing_names,
                    verbose=verbose, fLOG=fLOG, mapping=mapping,
                    rename=rename, level=level + 1)
                if len(m) > 0:
                    att = make_attribute(att.name, g)
                    mod += len(m)
                else:
                    att = make_attribute(att.name, att.g)
            new_node.attribute.append(att)
        if mod > 0:
            if verbose > 2:
                fLOG("[onnx_inline_function-renam] %s add node %r(%r): %r -> %r" % (
                    "  " * level,
                    new_node.op_type, new_node.name,
                    new_node.input, new_node.output))
            nodes.append(new_node)
            modified_nodes.append(node)
        else:
            nodes.append(node)

    if len(modified_nodes) > 0:
        if verbose > 1:
            fLOG("[onnx_inline_function-graph] %s -1 graph=%d "
                 "len(modified_nodes)=%d" % (
                     "  " * level, id(graph), len(modified_nodes)))

        graph = make_graph(
            nodes, graph.name, inputs, outputs,
            init, doc_string=graph.doc_string,
            sparse_initializer=list(graph.sparse_initializer))
    elif not rename:
        # no modification, let's check the node hiding a functions
        new_nodes = []
        for node in nodes:
            nnodes, m = _onnx_inline_function_node(
                node, protos, existing_names, verbose, fLOG,
                level=level)
            if len(m) > 0:
                if verbose > 0:
                    fLOG("[onnx_inline_function-subgr] %s replaced node %r (%r) "
                         "with %d nodes (id=%r) -- %r -> %r" % (
                             "  " * level,
                             node.name, node.op_type, len(nnodes), id(node),
                             node.input, node.output))
                new_nodes.extend(nnodes)
                modified_nodes.extend(m)
            else:
                new_nodes.append(node)
        if len(modified_nodes) > 0:
            if verbose > 1:
                fLOG("[onnx_inline_function-graph] %s -2 graph=%d "
                     "len(modified_nodes)=%d" % (
                         "  " * level, id(graph), len(modified_nodes)))

            nodes = new_nodes
            graph = make_graph(
                nodes, graph.name, inputs, outputs,
                init, doc_string=graph.doc_string,
                sparse_initializer=list(graph.sparse_initializer))

    if verbose > 1:
        fLOG("[onnx_inline_function-graph] %s <visit graph=%d end "
             "changed=%r len(modified_nodes)=%d" % (
                 "  " * level, id(graph0), id(graph0) != id(graph),
                 len(modified_nodes)))

    return graph, modified_nodes


def _onnx_inline_function_node(node, protos, existing_names, verbose,
                               fLOG, level):
    # The function does not rename input or output
    # of the node, it just replaces the node but a function
    # if the function exists.
    modified_nodes = []
    key = node.domain, node.op_type
    if key in protos:
        proto = protos[key]
        if not isinstance(proto, FunctionProto):
            raise TypeError(  # pragma: no cover
                "Prototype for key=%r must be a Function Proto, not %r." % (
                    key, type(proto)))
        modified_nodes.append(node)
        new_nodes = []
        mapping = _inline_mapping(verbose, fLOG, level)
        prefix = "_inl"

        for fr, to in zip(node.input, proto.input):
            n = make_node('Identity', [fr],
                          [_get_new_name(prefix, to, existing_names)])
            if verbose > 2:
                fLOG("[onnx_inline_function-ninpu] %s add node %r(%r): %r -> %r" % (
                    "  " * level, n.op_type, n.name, n.input, n.output))
            mapping[to] = n.output[0]
            if to != n.output[0]:
                mapping[n.output[0]] = n.output[0]
            new_nodes.append(n)

        for nn in proto.node:
            new_input = [mapping[i] for i in nn.input]
            new_output = [_get_new_name(prefix, o, existing_names)
                          for o in nn.output]
            mapping.update(
                {o: oo for o, oo in zip(nn.output, new_output)})
            mapping.update({oo: oo for oo in new_output})
            new_node = make_node(
                nn.op_type, new_input, new_output,
                domain=nn.domain, name=_get_new_name(
                    prefix, nn.name, existing_names))
            if verbose > 3:
                fLOG("[onnx_inline_function-nnode] %s rep node %r(%r): %r -> %r" % (
                    "  " * level, nn.op_type, nn.name, nn.input, nn.output))
            if verbose > 2:
                fLOG("[onnx_inline_function-nnode] %s add node %r(%r): %r -> %r" % (
                    "  " * level,
                    new_node.op_type, new_node.name,
                    new_node.input, new_node.output))
            for att in nn.attribute:
                if (att.type == AttributeProto.GRAPH and
                        hasattr(att, 'g') and att.g is not None):
                    if verbose > 1:
                        fLOG("[onnx_inline_function-funct] %s fct=%r graph=%d node=%d" % (
                            "  " * level, key, id(att.g), id(new_node)))

                    g, m = _onnx_inline_function_graph(
                        att.g, protos, existing_names=existing_names,
                        verbose=verbose, fLOG=fLOG, mapping=mapping,
                        rename=True, level=level + 1)
                    if len(m) > 0:
                        att = make_attribute(att.name, g)
                    else:
                        att = make_attribute(att.name, att.g)
                new_node.attribute.append(att)
            new_nodes.append(new_node)

        for fr, to in zip(proto.output, node.output):
            n = make_node('Identity', [mapping[fr]], [to])
            if verbose > 2:
                fLOG("[onnx_inline_function-noutt] %s add node %r(%r): %r -> %r" % (
                    "  " * level, n.op_type, n.name, n.input, n.output))
            new_nodes.append(n)
    else:
        new_nodes = [node]
        modified_nodes = []
    return new_nodes, modified_nodes


def onnx_inline_function(obj, protos=None, existing_names=None, verbose=0, fLOG=None):
    """
    Inlines functions in an ONNX graph.

    :param obj: onnx graph, :epkg:`FunctionProto`, :epkg:`GraphProto`,
        :epkg:`ModelProto`
    :param protos: if None, the function assumes *obj* is of type
        :epkg:`ModelProto` and the goal is to inline every function.
        If *protos* a list of strings, the function only inlines the
        functions in that list. If *protos* is a dictionary
        `{ (domain, type): FunctionProto }`, the function replaces every
        node `(domain, type)` by the code given in this dictionary
    :param existing_names: no new name will be taken in that set
    :param verbose: verbosity
    :param fLOG: logging function
    :return: modified object, list of modified nodes

    .. versionadded:: 0.9
    """
    if verbose > 0 and fLOG is None:
        fLOG = print
    if isinstance(obj, ModelProto):
        if verbose > 0:
            fLOG("[onnx_inline_function] type=%r graph=%d" % (
                type(obj), id(obj)))
        if protos is None:
            fct = [f.name for f in obj.functions]
            ex_names = set(enumerate_onnx_names(obj))
            if existing_names is not None:
                ex_names |= existing_names
            return onnx_inline_function(obj, fct, existing_names=ex_names,
                                        verbose=verbose, fLOG=fLOG)
        if isinstance(protos, list):
            ex_names = set(enumerate_onnx_names(obj))
            if existing_names is not None:
                ex_names |= existing_names
            protos = {(f.domain, f.name): f for f in obj.functions}
            return onnx_inline_function(obj, protos, existing_names=ex_names,
                                        verbose=verbose, fLOG=fLOG)
    if isinstance(protos, list):
        protos = {(f.domain, f.name): f for f in protos}
    if not isinstance(protos, dict):
        raise TypeError(  # pragma: no cover
            "obj is of type %r and protos must be a dictionary not %r." % (
                type(obj), type(protos)))

    if isinstance(obj, ModelProto):
        new_graph, m = onnx_inline_function(
            obj.graph, protos, verbose=verbose, fLOG=fLOG)
        if len(new_graph.initializer) != len(obj.graph.initializer):
            raise RuntimeError(  # pragma: no cover
                "Mismatched number of initializers %d != %d." % (
                    len(new_graph.initializer), len(obj.graph.initializer)))
        if len(new_graph.sparse_initializer) != len(obj.graph.sparse_initializer):
            raise RuntimeError(  # pragma: no cover
                "Mismatched number of initializers %d != %d." % (
                    len(new_graph.sparse_initializer),
                    len(obj.graph.sparse_initializer)))
        new_functions = []
        distri = Counter(
            (n.domain, n.op_type)
            for n in enumerate_onnx_nodes(new_graph))
        opsets = {op.domain: op.version for op in obj.opset_import}
        for f in obj.functions:
            key = f.domain, f.name
            if key not in protos:
                new_functions.append(f)
            elif key in distri:
                raise RuntimeError(  # pragma: no cover
                    "Function %r still appears in the graph, "
                    "distibution=%s." % (key, pprint.pformat(distri)))
            if f.domain not in opsets:
                opsets[f.domain] = 1
        return (
            make_model(
                new_graph,
                functions=new_functions,
                opset_imports=[
                    make_operatorsetid(k, v)
                    for k, v in opsets.items()],
                producer_name=obj.producer_name,
                producer_version=obj.producer_version,
                ir_version=obj.ir_version,
                doc_string=obj.doc_string,
                domain=obj.domain,
                model_version=obj.model_version),
            m)

    # FunctionProto, GraphProto
    if existing_names is None:
        existing_names = set(enumerate_onnx_names(obj))

    if verbose > 0:
        fLOG("[onnx_inline_function] type=%r graph=%d begin" % (
            type(obj), id(obj)))
        distri = Counter((n.domain, n.op_type)
                         for n in enumerate_onnx_nodes(obj))

    new_nodes = list(obj.node)
    modified_nodes = []
    n_iter = 0
    max_iter = onnx_subgraphs_level(obj) + 1
    modified = 1
    while modified > 0 and n_iter < max_iter:
        if verbose > 0:
            fLOG(f"[onnx_inline_function] start iteration {n_iter!r}")

        # local context
        mapping = _inline_mapping(verbose, fLOG, level=0)
        if isinstance(obj, GraphProto):
            mapping.update({i.name: i.name for i in obj.initializer})
            mapping.update({i.name: i.name for i in obj.sparse_initializer})
            for i in obj.input:
                if i.name not in mapping:
                    mapping[i.name] = i.name
        elif isinstance(obj, FunctionProto):
            mapping.update({i: i for i in obj.input})
        else:
            raise TypeError(  # pragma: no cover
                f"Unexpected type for obj: {type(obj)!r}.")

        # loop on nodes
        old_nodes = new_nodes
        modified = 0
        new_nodes = []
        for node in old_nodes:
            nnodes, m = _onnx_inline_function_node(
                node, protos, existing_names, verbose, fLOG, level=0)
            mapping.update({o: o for o in node.output})

            if len(m) > 0:
                if verbose > 0:
                    fLOG("[onnx_inline_function] replaced node %r (%r) "
                         "with %d nodes (id=%r) -- %r -> %r (iter=%r)" % (
                             node.name, node.op_type, len(nnodes), id(node),
                             node.input, node.output, n_iter))
                modified += len(m)
                new_nodes.extend(nnodes)
                modified_nodes.extend(m)
            else:
                has_graph = False
                new_attributes = []
                for att in node.attribute:
                    if (att.type == AttributeProto.GRAPH and
                            hasattr(att, 'g') and att.g is not None):
                        g, m = _onnx_inline_function_graph(
                            att.g, protos, verbose=verbose, fLOG=fLOG,
                            existing_names=existing_names, mapping=mapping,
                            rename=False, level=1)
                        if len(m) > 0:
                            modified_nodes.extend(m)
                            modified_nodes.append(node)
                            modified += 1 + len(m)
                            has_graph = True
                            att = make_attribute(att.name, g)
                    new_attributes.append(att)
                if has_graph:
                    new_node = make_node(
                        node.op_type, node.input, node.output,
                        domain=node.domain, name=node.name)
                    new_node.attribute.extend(new_attributes)
                    new_nodes.append(new_node)
                else:
                    # we still need to check that this subgraph does
                    # not include a function
                    new_nodes.append(node)

        n_iter += 1
        if verbose > 0:
            total_node = len(list(enumerate_onnx_nodes(new_nodes)))
            fLOG("[onnx_inline_function] n_iter=%r/%r nodes=%r modified=%r "
                 "n_nodes=%d total=%d" % (
                     n_iter, max_iter, len(obj.node), modified,
                     len(new_nodes), total_node))

    if verbose > 0:
        fLOG("[onnx_inline_function] type=%r graph=%d end with %d "
             "modified nodes" % (
                 type(obj), id(obj), len(modified_nodes)))
        distri2 = Counter((n.domain, n.op_type)
                          for n in enumerate_onnx_nodes(new_nodes))
        if distri != distri2:
            fLOG("[onnx_inline_function] BEFORE")
            for k, v in sorted(distri.items()):
                fLOG("[onnx_inline_function] %d -- %s" % (v, k))
            fLOG("[onnx_inline_function] AFTER")
            for k, v in sorted(distri2.items()):
                fLOG("[onnx_inline_function] %d -- %s" % (v, k))

    if isinstance(obj, FunctionProto):
        return (
            make_function(
                domain=obj.domain, fname=obj.name,
                inputs=obj.input, outputs=obj.output, nodes=new_nodes,
                opset_imports=[
                    make_operatorsetid(op.domain, op.version)
                    for op in obj.opset_import],
                doc_string=obj.doc_string,
                attributes=obj.attribute),
            modified_nodes)
    if isinstance(obj, GraphProto):
        return (
            make_graph(new_nodes, obj.name, list(obj.input), list(obj.output),
                       list(obj.initializer), doc_string=obj.doc_string,
                       sparse_initializer=list(obj.sparse_initializer)),
            modified_nodes)
    raise TypeError(  # pragma: no cover
        f"Unexpected type for obj {type(obj)!r}.")



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Transpose(OpRunUnaryNum):

    atts = {'perm': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Transpose.atts,
                               **options)
        self.perm_ = None if len(self.perm) == 0 else self.perm

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.perm_ is None:
            return (numpy.transpose(data), )
        if len(self.perm_) != len(data.shape):
            raise RuntimeError(  # pragma: no cover
                f"Inconsistent permutation {self.perm_!r} with shape {data.shape!r}.")
        return (numpy.transpose(data, axes=self.perm_), )

    def to_python(self, inputs):
        """
        Returns a python code equivalent to this operator.

        @param      inputs      inputs name
        @return                 imports, python code, both as strings
        """
        lines = [
            "if perm is None:",
            f"    return numpy.transpose({inputs[0]})",
            f"return numpy.transpose({inputs[0]}, axes=perm)"
        ]
        return "import numpy", "\n".join(lines)
